"""
    Declare `PluginMount` and various extension points

    To define a Plugin: 

    1.  subclass from the desired extension point class
    2.  define a class method `register` that fills the declares
        the relevant attributes by calling `add_argument`
        of `cls.schema`
    3. define a `plugin_name` class attribute

    To define an ExtensionPoint:

    1. add a class decorator @ExtensionPoint
"""
from nbodykit.utils.config import autoassign, ConstructorSchema, ReadConfigFile, PluginParsingError
from nbodykit.distributedarray import ScatterArray

import numpy
from argparse import Namespace
import functools
import weakref


# MPI will be required because
# a plugin instance will be created for a MPI communicator.
from mpi4py import MPI

algorithms  = Namespace()
datasources = Namespace()
painters    = Namespace()
transfers   = Namespace()

# private variable to store global MPI communicator 
# that all plugins are initialized with
_comm = MPI.COMM_WORLD
_cosmo = None

def get_nbkit_comm():
    """
    Return the global MPI communicator that all plugins 
    will be instantiated with (stored in `comm` attribute of plugin)
    """
    return _comm
    
def set_nbkit_comm(comm):
    """
    Set the global MPI communicator that all plugins 
    will be instantiated with (stored in `comm` attribute of plugin)
    """
    global _comm
    _comm = comm
    
def get_nbkit_cosmo():
    """
    Return the global Cosmology instance
    """
    return _cosmo
    
def set_nbkit_cosmo(cosmo):
    """
    Set the global Cosmology instance
    """
    global _cosmo
    _cosmo = cosmo

class PluginInterface(object):
    """ 
    The basic interface of a plugin -- classes must implement
    the 'register' function
    """
    @classmethod
    def register(kls):
        raise NotImplementedError

        
def ExtensionPoint(registry):
    """ 
    Declares a class as an extension point, registering to registry 
    """
    def wrapped(cls):
        cls = add_metaclass(PluginMount)(cls)
        cls.registry = registry
        return cls
    return wrapped

class PluginMount(type):
    """ 
    Metaclass for an extension point that provides the 
    methods to manage plugins attached to the extension point
    """
    def __new__(cls, name, bases, attrs):
        # Add PluginInterface to the ExtensionPoint,
        # Plugins at an extensioni point will inherit from PluginInterface
        # This is more twisted than it could have been!

        if len(bases) == 0 or (len(bases) == 1 and bases[0] is object):
            bases = (PluginInterface,)
        return type.__new__(cls, name, bases, attrs)

    def __init__(cls, name, bases, attrs):
        # only executes when processing the mount point itself.
        # the extension mount point only declares a PluginInterface
        # the plugins at an extension point will always be its subclass
        
        if cls.__bases__ == (PluginInterface, ):
            return

        if not hasattr(cls, 'plugin_name'):
            raise RuntimeError("Plugin class must carry a 'plugin_name'")
        
        # register, if this plugin isn't yet
        if cls.plugin_name not in cls.registry:

            # initialize the schema and alias it
            if cls.__init__ == object.__init__:
                raise ValueError("please define an __init__ method for '%s'" %cls.__name__)
            
            # in python 2, __func__ needed to attach attributes to the real function; 
            # __func__ removed in python 3, so just attach to the function
            init = cls.__init__
            if hasattr(init, '__func__'):
                init = init.__func__

            # add a schema
            init.schema = ConstructorSchema()
            cls.schema = cls.__init__.schema

            # track names of classes
            setattr(cls.registry, cls.plugin_name, cls)

            # register the class
            cls.register()

            # configure the class __init__, attaching the comm, and optionally cosmo
            attach_cosmo = issubclass(cls, DataSource)
            cls.__init__ = autoassign(init, attach_cosmo=attach_cosmo)

    def create(cls, plugin_name, use_schema=False, **kwargs):
        """
        Instantiate a plugin from this extension point,
        based on the name/value pairs passed as keywords.

        Optionally, cast the keywords values, using the types
        defined by the schema of the class we are creating

        Parameters
        ----------
        plugin_name: str
            the name of the plugin to instantiate
        use_schema : bool, optional
            if `True`, cast the kwargs that are defined in 
            the class schema before initializing. Default: `False`
        kwargs : (key, value) pairs
            the parameter names and values that will be
            passed to the plugin's __init__

        Returns
        -------
        plugin : 
            the initialized instance of `plugin_name`
        """
        if plugin_name not in cls.registry:
            raise ValueError("'%s' does not match the names of any loaded plugins" %plugin_name)
            
        klass = getattr(cls.registry, plugin_name)
        
        # cast the input values, using the class schema
        if use_schema:
            for k in kwargs:
                if k in klass.schema:
                    arg = klass.schema[k]
                    kwargs[k] = klass.schema.cast(arg, kwargs[k])
                        
        toret = klass(**kwargs)
        
        ### FIXME: not always using create!!
        toret.string = id(toret)
        return toret
        
    def from_config(cls, parsed): 
        """ 
        Instantiate a plugin from this extension point,
        based on the input `parsed` value, which is parsed
        directly from the YAML configuration file
        
        There are several valid input cases for `parsed`:
            1.  parsed: dict
                containing the key `plugin`, which gives the name of 
                the Plugin to load; the rest of the dictionary is 
                treated as arguments of the Plugin
            2.  parsed: dict
                having only one entry, with key giving the Plugin name
                and value being a dictionary of arguments of the Plugin
            3.  parsed: dict
                if `from_config` is called directly from a Plugin class, 
                then `parsed` can be a dictionary of the named arguments,
                with the Plugin name inferred from the class `cls`
            4.  parsed: str
                the name of a Plugin, which will be created with 
                no arguments
        """    
        try:    
            if isinstance(parsed, dict):
                if 'plugin' in parsed:
                    kwargs = parsed.copy()
                    plugin_name = kwargs.pop('plugin')
                    return cls.create(plugin_name, use_schema=True, **kwargs)
                elif len(parsed) == 1:
                    k = list(parsed.keys())[0]
                    if isinstance(parsed[k], dict):
                        return cls.create(k, use_schema=True, **parsed[k])
                    else:
                        raise PluginParsingError
                elif hasattr(cls, 'plugin_name'):
                    return cls.create(cls.plugin_name, use_schema=True, **parsed)
                else:
                    raise PluginParsingError
            elif isinstance(parsed, str):
                return cls.create(parsed)
            else:
                raise PluginParsingError
        except PluginParsingError as e:
            msg = '\n' + '-'*75 + '\n'
            msg += "failure to parse plugin from configuration using `from_config()`\n"
            msg += ("\nThere are several ways to initialize plugins from configuration files:\n"
                    "1. The plugin parameters are specified as a dictionary containing the key `plugin`,\n"
                    "   which gives the name of the plugin to load; the rest of the dictionary is\n"
                    "   passed to the plugin `__init__()` as keyword arguments\n"
                    "2. The plugin is specified as a dictionary having only one entry -- \n"
                    "   the key gives the plugin name and the value is a dict of arguments\n"
                    "   to be passed to the plugin `__init__()`\n"
                    "3. When `from_config()` is bound to a particular plugin class, only a dict\n"
                    "   of the `__init__()` arguments should be specified\n"
                    "4. The plugin is specified as a string, which gives the name of the plugin;\n"
                    "   the plugin will be created with no arguments\n")
            msg += '\n' + '-'*75 + '\n'
            e.args = (msg,)
            raise 
        except:
            raise
            
    def format_help(cls, *plugins):
        """
        Return a string specifying the `help` for each of the plugins
        specified, or all if none are specified
        """
        if not len(plugins):
            plugins = list(vars(cls.registry).keys())
            
        s = []
        for k in plugins:
            if not isplugin(k):
                raise ValueError("'%s' is not a valid Plugin name" %k)
            header = "Plugin : %s  ExtensionPoint : %s" % (k, cls.__name__)
            s.append(header)
            s.append("=" * (len(header)))
            s.append(getattr(cls.registry, k).schema.format_help())

        if not len(s):
            return "No available Plugins registered at %s" %cls.__name__
        else:
            return '\n'.join(s) + '\n'

# copied from six
def add_metaclass(metaclass):
    """Class decorator for creating a class with a metaclass."""
    def wrapper(cls):
        orig_vars = cls.__dict__.copy()
        slots = orig_vars.get('__slots__')
        if slots is not None:
            if isinstance(slots, str):
                slots = [slots]
            for slots_var in slots:
                orig_vars.pop(slots_var)
        orig_vars.pop('__dict__', None)
        orig_vars.pop('__weakref__', None)
        return metaclass(cls.__name__, cls.__bases__, orig_vars)
    return wrapper

@ExtensionPoint(transfers)
class Transfer:
    """
    Mount point for plugins which apply a k-space transfer function
    to the Fourier transfrom of a datasource field
    
    Plugins implementing this reference should provide the following 
    attributes:

    plugin_name : str
        class attribute giving the name of the subparser which 
        defines the necessary command line arguments for the plugin
    
    register : classmethod
        A class method taking no arguments that adds a subparser
        and the necessary command line arguments for the plugin
    
    __call__ : method
        function that will apply the transfer function to the complex array
    """
    def __call__(self, pm, complex):
        """ 
        Apply the transfer function to the complex field
        
        Parameters
        ----------
        pm : ParticleMesh
            the particle mesh object which holds possibly useful
            information, i.e, `w` or `k` arrays
        complex : array_like
            the complex array to apply the transfer to
        """
        raise NotImplementedError

class DataStream(object):
    """
    A class to represent an open `DataSource` stream, which
    is called by `DataSource.open()`
    
    The class is written such that it be used with or without
    the `with` statement, similar to the file `open()` function
    
    
    Attributes
    ----------
    read : callable
        a method that returns an iterator that will iterate through
        the data source (either collectively or uncollectively), 
        returning the specified columns
    nread : int
        during each iteration of `read`, this will store the number
        of rows read from the data source
    """
    def __init__ (self, data, defaults={}):
        
        # store the data and defaults
        self.data = data
        self.defaults = defaults
        self.nread = 0
        
        # store a weak reference to the private cache
        self._cacheref = data._cache
        if not self._cacheref.empty:
            self.nread = self._cacheref.total_size
            
    def __enter__ (self):
        return self
        
    def __exit__ (self, exc_type, exc_value, traceback):
        self.close()
        
    def close(self):
        """
        Close the DataStream; no `read` operations are
        allowed on closed streams
        """
        if not self.closed:
            del self._cacheref
        
    @property
    def closed(self):
        """
        Whether or not the DataStream is closed
        """
        return not hasattr(self, '_cacheref')        
        
    def isdefault(self, name, data):
        """
        Return `True` if the input data is equal to the default
        value for this stream
        """
        if name in self.defaults:
            return self.defaults[name] == data[0]
        else:
            return False
        
    def read(self, columns, full=False):
        """
        Read the data corresponding to `columns`
        
        Parameters
        ----------
        columns : list
            list of strings giving the names of the desired columns
        full : bool, optional
            if `True`, any `bunchsize` parameters will be ignored, so 
            that each rank will read all of its specified data section
            at once; default is `False`
        
        Returns
        -------
        data : list
            a list of the data for each column in `columns`
        """
        if self.closed:
            raise ValueError("'read' operation on closed data stream")
            
        # return data from cache, if it's not empty
        if not self._cacheref.empty:
            
            # valid column names
            valid_columns = list(set(self._cacheref.columns)|set(self.defaults))
            
            # replace any missing columns with None
            data = []
            for i, col in enumerate(columns):
                if col in self._cacheref.columns:
                    data.append(self._cacheref[col])
                else:
                    data.append(None)
            
            # yield the blended data
            yield self._blend_data(columns, data, valid_columns, self._cacheref.local_size)
            
        # parallel read is defined
        else:
            # reset the number of rows read
            self.nread = 0
            
            # do the parallel read
            for data in self.data.parallel_read(columns, full=full):
                
                # determine the valid columns (where data is not None)
                valid_columns = []; indices = []
                for i, col in enumerate(columns):
                    if data[i] is not None:
                        valid_columns.append(col)
                        indices.append(i)
                valid_columns = list(set(valid_columns)|set(self.defaults))
                
                # verify data
                valid_data = [data[i] for i in indices]
                size = self.data._verify_data(valid_data)
                
                # update the number of rows read
                self.nread += self.data.comm.allreduce(size)
                
                # yield the blended data with defaults
                yield self._blend_data(columns, data, valid_columns, size)
                
    def _blend_data(self, columns, data, valid_columns, size):
        """
        Internal function to blend data that has been explicitly read
        and any default values that have been set. 
        
        Notes
        -----
        *   Default values are returned with the correct size but minimal
            memory usage using `stride_tricks` from `numpy.lib`
        *   This function will crash if a column is requested that the 
            data source does not provided and no default exists
        
        Parameters
        ----------
        columns : list
            list of the column names that are being returned
        data : list
            list of data corresponding to `columns`; if a column is not 
            supported, the element of data is `None`
        valid_columns : list
            the list of valid column names; the union of the columns supported
            by the data source and the default values
        size : int
            the size of the data we are returning
        
        Returns
        -------
        newdata : list
            the list of data with defaults blended in, if need be
        """
        newdata = []
        for i, col in enumerate(columns):
                        
            # this column is missing -- crash
            if col not in valid_columns:
                args = (col, str(valid_columns))
                raise DataSource.MissingColumn("column '%s' is unavailable; valid columns are: %s" %args)
            
            # we have this column
            else:
                # return read data, if not None
                if data[i] is not None:
                    newdata.append(data[i])
                # return the default
                else:
                    if col not in self.defaults:
                        raise RuntimeError("missing default value when trying to blend data")    
                        
                    # use stride_tricks to avoid memory usage
                    val = numpy.asarray(self.defaults[col])
                    d = numpy.lib.stride_tricks.as_strided(val, (size, val.size), (0, val.itemsize))
                    newdata.append(numpy.squeeze(d))
            
        return newdata
        
class DataCache(object):
    """
    A class to cache data in manner that can be weakly
    referenced via `weakref`
    """
    def __init__(self, columns, data, local_size, total_size):
        
        self.columns    = columns
        self.local_size = local_size
        self.total_size = total_size
    
        for col, d in zip(columns, data):
            setattr(self, col, d)
    
    @property
    def empty(self):
        return len(self.columns) == 0
        
    def __getitem__(self, col):
        if col in self.columns:
            return getattr(self, col)
        else:
            raise KeyError("no such column in DataCache; available columns: %s" %str(self.columns))    
    
    def __str__(self):
        return self.__repr__()
        
    def __repr__(self):
        return 'DataCache(%s)' %str(self.columns)
         
@ExtensionPoint(datasources)
class DataSource:
    """
    Mount point for plugins which refer to the reading of input files.
    The `read` operation occurs on a `DataStream` object, which
    is returned by `DataSource.open()`
    
    Default values for any columns to read can be supplied as a 
    dictionary argument to `DataSource.open()`
    
    Notes
    -----
    *   a `Cosmology` instance can be passed to any `DataSource`
        class via the `cosmo` keyword
    *   the data will be cached in memory if returned via `readall`;
        the default behavior is for the cache to persist while 
        an open `DataStream` is valid, but the cache can be forced
        to persist via the `keep_cache` context manager

    Plugins implementing this reference should provide the following 
    attributes:

    plugin_name : str
        class attribute that defines the name of the Plugin in 
        the registry
    
    register : classmethod
        A class method taking no arguments that declares the
        relevant attributes for the class by adding them to the
        class schema
    
    readall: method
        A method to read all available data at once (uncollectively) 
        and cache the data in memory for repeated calls to `read`

    parallel_read: method
        A method to read data for complex, large data sets. The read
        operation shall be collective, with each yield generating 
        different sections of the datasource on different ranks. 
        No caching of data takes places.
    """
    class MissingColumn(Exception):
        pass
            
    def _cache_data(self):
        """
        Internal function to cache the data from `readall`.
        
        This function performs the following actions:
        
             1. calls `readall` on the root rank
             2. scatters read data evenly across all ranks
             3. stores the scattered data as the `cache` attribute
             4. stores the total size of the data as the `size` attribute
             5. stores the available columns in the cache as `cached_columns`
             
        Returns
        -------
        success : bool
            if the call to `readall` is successful, returns `True`, else `False`
        """
        # rank 0 tries to readall, and tells everyone else if it succeeds
        success = False
        if self.comm.rank == 0:
            
            # read all available data
            try:
                data = self.readall()
                success = True
            except NotImplementedError:
                pass
            except Exception:
                import traceback
                raise Exception("an unknown exception occurred while trying to cache data via `readall`:\n"
                                "traceback:\n\n%s" %traceback.format_exc())
                
        # determine if readall was successful
        success = self.comm.bcast(success, root=0)
    
        # cache the data
        if success:
            
            columns = []; size = None
            if self.comm.rank == 0:
                columns = list(data.keys())
                data = [data[c] for c in columns] # store a list
            
            # everyone gets the column names
            columns = self.comm.bcast(columns, root=0)
        
            # verify the input data
            if self.comm.rank == 0:    
                size = self._verify_data(data)
            else:
                data = [None for c in columns]
        
            # everyone gets the total size
            size = self.comm.bcast(size, root=0)
        
            # scatter the data across all ranks
            # each rank caches only a part of the total data
            cache = []; local_sizes = []
            for d in data:
                cache.append(ScatterArray(d, self.comm, root=0))
                local_sizes.append(len(cache[-1]))
                
            # this should hopefully never fail (guaranted to have nonzero length)
            if not all(s == local_sizes[0] for s in local_sizes):
                raise RuntimeError("scattering data resulted in uneven lengths between columns")
            local_size = local_sizes[0]

            # the total collective size of the datasource
            self.size = size # this will persist, even if cache is deleted
            
            # return the cache
            return DataCache(columns, cache, local_size, size)
        else:
            return DataCache([], [], 0, 0) # empty cache

    def _verify_data(self, data):
        """
        Internal function to verify the input data by checking that: 
            
            1. `data` is not empty
            2. the sizes of each element of data are equal
        
        Parameters
        ----------
        data : list
            the list of data arrays corresponding to the requested columns
        
        Returns
        -------
        size : int
            the size of each element of `data`
        """
        # need some data
        if not len(data):
            raise ValueError('DataSource::read did not return any data')
        
        # make sure the number of rows in each column read is equal
        # columns has to have length >= 1, or we crashed already
        if not all(len(d) == len(data[0]) for d in data):
            raise RuntimeError("column length mismatch in DataSource::read")
        
        # return the size
        return len(data[0])
                    
    @property
    def _cache(self):
        """
        Internal cache property storing a `DataCache`. This is designed to 
        be accessed only by `DataStream` objects
        """
        # create the weakref dict if need be
        if not hasattr(self, '_weakcache'):
            self._weakcache = weakref.WeakValueDictionary()
            
        # create the DataCache if need be
        if 'cache' not in self._weakcache:
            cache = self._cache_data()
            self._weakcache['cache'] = cache
            
        return self._weakcache['cache']
        
    @staticmethod
    def BoxSizeParser(value):
        """
        Read the `BoxSize, enforcing that the BoxSize must be a 
        scalar or 3-vector
        
        Returns
        -------
        BoxSize : array_like
            an array of size 3 giving the box size in each dimension
        """
        boxsize = numpy.empty(3, dtype='f8')
        try:
            if isinstance(value, (tuple, list)) and len(value) != 3:
                raise ValueError
            boxsize[:] = value
        except:
            raise ValueError("DataSource `BoxSize` must be a scalar or three-vector")
        return boxsize
    
    @property
    def size(self):
        """
        The total size of the DataSource returned via the `read` operation.
        
        The user can set this explicitly if the size is known before 
        `read` is called, otherwise it will be set (exactly once) after `read` is called 
        """
        try:
            return self._size
        except:
            raise AttributeError("DataSource `size` is not known a priori (i.e., before the 'read' operation)")
    
    @size.setter
    def size(self, val):
        """
        Set the `size` attribute. This should only be set once
        """
        if not hasattr(self, '_size'):
            self._size = val
        else:
            if val != self._size:
                raise ValueError("DataSource `size` has already been set to a different value")
    
    def keep_cache(self):
        """
        A context manager that forces the `DataSource` cache to persist,
        even if there are no open `DataStream` objects. This will 
        prevent unwanted and unnecessary rereading of the `DataSource`

        The below example details the intended usage. The data is 
        only cached once below:
        
            >> with datasource.keep_cache():
            >>    with datasource.open() as stream1:
            >>      [[pos]] = stream1.read(['Position'], full=True)
            >>    
            >>    with datasource.open() as stream2:
            >>      [[vel]] = stream2.read(['Velocity'], full=True)
        """
        # simplest implementation is returning a stream
        return DataStream(self) 
    
    def open(self, defaults={}):
        """
        Open the `DataSource` by returning a `DataStream` from which
        the data can be read. 
        
        This function also specifies the default values for any columns
        that are not supported by the `DataSource`. The defaults are
        unique to each `DataStream`, but a DataSource can be opened
        multiple times (returning different streams) with different 
        default values
        
        Parameters
        ----------
        defaults : dict, optional
            a dictionary providing default values for a given column
        name : str, optional
            return a specific named stream; the datasource must have
            an attribute with this name, which is a `DataSource` 
        
        Returns
        -------
        stream : DataStream
            the stream object from which the data can be read via
            the `read` function
        """
        # note: if DataSource is already `open`, we can 
        # still get a new stream with different defaults
        return DataStream(self, defaults=defaults)
                             
    def readall(self):
        """ 
        Override to provide a method to read all available data at once 
        (uncollectively) and cache the data in memory for repeated 
        calls to `read`

        Notes
        -----
        *   The default 'read' function calls this function on the
            root rank to read all available data, and then scatters
            the data evenly across all available ranks
        *   The intention is to reduce the complexity of implementing a 
            simple and small data source, for which reading all data at once
            is feasible
            
        Returns
        -------
        data : dict
            a dictionary of all supported data for the datasource; keys
            give the column names and values are numpy arrays
        """
        raise NotImplementedError
        
    def parallel_read(self, columns, full=False):
        """ 
        Override this function for complex, large data sets. The read
        operation shall be collective, each yield generates different
        sections of the datasource. No caching of data takes places.
        
        If the `DataSource` does not provide a column in `columns`, 
        `None` should be returned
        
        Notes
        -----
        *   This function will be called if `readall` is not provided
        *   The intention is for this function to handle complex and 
            large data sets, where parallel I/O across ranks are 
            required to avoid memory and I/O issues
            
        Parameters
        ----------
        full : bool, optional
            if `True`, any `bunchsize` parameters will be ignored, so 
            that each rank will read all of its specified data section
            at once
        
        Returns
        -------
        data : list
            a list of the data for each column in columns; if the datasource
            does not provide a given column, that element should be `None`
        """
        raise NotImplementedError
        

@ExtensionPoint(painters)
class Painter:
    """
    Mount point for plugins which refer to the painting of input files.

    Plugins implementing this reference should provide the following 
    attributes:

    plugin_name : str
        class attribute giving the name of the subparser which 
        defines the necessary command line arguments for the plugin
    
    register : classmethod
        A class method taking no arguments that adds a subparser
        and the necessary command line arguments for the plugin
    
    paint : method
        A method that performs the painting of the field.

    """
    
    def paint(self, pm, datasource):
        """ 
            Paint from a data source. 

            It shall read data then use basepaint to paint.
            
            Returns a dictionary of attributes of the painting operation.
        """
        raise NotImplementedError

    def basepaint(self, pm, position, weight=None):
        assert pm.comm == self.comm # pm must be from the same communicator!
        layout = pm.decompose(position)
        Nlocal = len(position)
        position = layout.exchange(position)
        if weight is not None:
            weight = layout.exchange(weight)
            pm.paint(position, weight)
        else:
            pm.paint(position) 
        return Nlocal 
        
@ExtensionPoint(algorithms)
class Algorithm:
    """
    Mount point for plugins which provide an interface for running
    one of the high-level algorithms, i.e, power spectrum calculation
    or FOF halo finder
    
    Plugins implementing this reference should provide the following 
    attributes:

    plugin_name : str
        class attribute giving the name of the subparser which 
        defines the necessary command line arguments for the plugin
    
    register : classmethod
        A class method taking no arguments that adds a subparser
        and the necessary command line arguments for the plugin
    
    __call__ : method
        function that will apply the transfer function to the complex array
    """        
    def run(self):
        raise NotImplementedError
    
    def save(self, *args, **kwargs):
        raise NotImplementedError
    
    @classmethod
    def parse_known_yaml(kls, name, stream):
        """
        Parse the known (and unknown) attributes from a YAML, where `known`
        arguments must be part of the Algorithm.parser instance
        """
        # get the class for this algorithm name
        klass = getattr(kls.registry, name)
        
        # get the namespace from the config file
        return ReadConfigFile(stream, klass.schema)


__valid__ = [DataSource, Painter, Transfer, Algorithm]
__all__ = list(map(str, __valid__))

def isplugin(name):
    """
    Return `True`, if `name` is a registered plugin for any extension point
    """
    for extensionpt in __valid__:
        if name in extensionpt.registry: return True
    
    return False



from nbodykit.extensionpoints import DataSource
import numpy
import logging

logger = logging.getLogger('UniformBox')
         
class UniformBoxDataSource(DataSource):
    """
    DataSource to return the following columns:
    
    `Position` : uniformly distributed in box
    `Velocity` : uniformly distributed between ``+/- max_speed``
    `LogMass`  : normally distributed with mean `mu_logM` and std dev `sigma_logM`
    `Mass`     : values corresponding to 10**`LogMass`
    """
    plugin_name = "UniformBox"
    
    def __init__(self, N, BoxSize, max_speed=500., mu_logM=13.5, sigma_logM=0.5, seed=None):        
        
        # initalize a random state for each rank
        if self.seed is not None:
            seed = self.seed*(self.comm.rank+1)
            self.random = numpy.random.RandomState(seed)
        else:
            self.random = numpy.random
            
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "data particles with uniform positions and velocities"
        
        s.add_argument("N", type=int,
            help='the total number of objects in the box')
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="the size of the isotropic box, or the sizes of the 3 box dimensions")
        s.add_argument("max_speed", type=float, 
            help="use uniform velocities between [-max_speed, max_speed] (in km/s)")
        s.add_argument("mu_logM", type=float, 
            help="the mean log10 mass to use when generating mass values")
        s.add_argument("sigma_logM", type=float, 
            help="the standard deviation of the log10 mass to use when generating mass values")
        s.add_argument("seed", type=int,
            help='the number used to seed the random number generator')
    
    def readall(self):
        """
        Valid columns are:
            `Position` : uniformly distributed in box
            `Velocity` : uniformly distributed between ``+/- max_speed``
            `LogMass`  : normally distributed with mean `mu_logM` and std dev `sigma_logM`
            `Mass`     : values corresponding to 10**`LogMass` 
        """            
        toret = {}
        shape = (self.N, 3)
        
        toret['Position'] = self.random.uniform(size=shape) * self.BoxSize
        toret['Velocity'] = 2*self.max_speed * self.random.uniform(size=shape) - self.max_speed
        toret['LogMass']  = self.random.normal(loc=self.mu_logM, scale=self.sigma_logM, size=self.N)
        toret['Mass']     = 10**(toret['LogMass'])
        
        return toret




import numpy
import mpsort
from mpi4py import MPI

def GatherArray(data, comm, root=0):
    """
    Gather the input data array from all ranks to the specified ``root``
    
    This uses `Gatherv`, which avoids mpi4py pickling, and also
    avoids the 2 GB mpi4py limit for bytes using a custom datatype
    
    Parameters
    ----------
    data : array_like
        the data on each rank to gather 
    comm : MPI communicator
        the MPI communicator
    root : int
        the rank number to gather the data to
        
    Returns
    -------
    recvbuffer : array_like, None
        the gathered data on root, and `None` otherwise
    """
    if not isinstance(data, numpy.ndarray): 
        raise ValueError("`data` must by numpy array in GatherArray")
        
    # need C-contiguous order
    if not data.flags['C_CONTIGUOUS']:
        data = numpy.ascontiguousarray(data)
    local_length = data.shape[0]
    
    # check dtypes and shapes
    shapes = comm.gather(data.shape, root=root)
    dtypes = comm.gather(data.dtype, root=root)
    
    if comm.rank == root:
        if any(s[1:] != shapes[0][1:] for s in shapes):
            raise ValueError("mismatch between shape[1:] across ranks in GatherArray")
        if any(dt != dtypes[0] for dt in dtypes):
            raise ValueError("mismatch between dtypes across ranks in GatherArray")
        
    shape = data.shape
    dtype = data.dtype
        
    # setup the custom dtype 
    duplicity = numpy.product(numpy.array(shape[1:], 'intp'))
    itemsize = duplicity * dtype.itemsize
    dt = MPI.BYTE.Create_contiguous(itemsize)
    dt.Commit()
        
    # compute the new shape for each rank
    newlength = comm.allreduce(local_length)
    newshape = list(shape)
    newshape[0] = newlength

    # the return array
    if comm.rank == root:
        recvbuffer = numpy.empty(newshape, dtype=dtype, order='C')
    else:
        recvbuffer = None

    # the recv counts
    counts = comm.allgather(local_length)
    counts = numpy.array(counts, order='C')
    
    # the recv offsets
    offsets = numpy.zeros_like(counts, order='C')
    offsets[1:] = counts.cumsum()[:-1]
    
    # gather to root
    comm.Barrier()
    comm.Gatherv([data, dt], [recvbuffer, (counts, offsets), dt], root=root)
    dt.Free()
    
    return recvbuffer
     
def ScatterArray(data, comm, root=0):
    """
    Scatter the input data array across all ranks, assuming `data` is 
    initially only on `root` (and `None` on other ranks)
    
    This uses `Scatterv`, which avoids mpi4py pickling, and also
    avoids the 2 GB mpi4py limit for bytes using a custom datatype
    
    Parameters
    ----------
    data : array_like or None
        on `root`, this gives the data to split and scatter 
    comm : MPI communicator
        the MPI communicator
    root : int
        the rank number that initially has the data
        
    Returns
    -------
    recvbuffer : array_like
        the chunk of `data` that each rank gets
    """
    if comm.rank == root:
        if not isinstance(data, numpy.ndarray): 
            raise ValueError("`data` must by numpy array on root in ScatterArray")
        
        # need C-contiguous order
        if not data.flags['C_CONTIGUOUS']:
            data = numpy.ascontiguousarray(data)
        shape_and_dtype = (data.shape, data.dtype)
    else:
        shape_and_dtype = None
        
    # each rank needs shape/dtype of input data
    shape, dtype = comm.bcast(shape_and_dtype)
    
    # initialize empty data on non-root ranks
    if comm.rank != root:
        data = numpy.empty(0, dtype=(dtype, shape[1:]))
    
    # setup the custom dtype 
    duplicity = numpy.product(numpy.array(shape[1:], 'intp'))
    itemsize = duplicity * dtype.itemsize
    dt = MPI.BYTE.Create_contiguous(itemsize)
    dt.Commit()
        
    # compute the new shape for each rank
    newshape = list(shape)
    newlength = shape[0] // comm.size
    if comm.rank < shape[0] % comm.size:
        newlength += 1
    newshape[0] = newlength

    # the return array
    recvbuffer = numpy.empty(newshape, dtype=dtype, order='C')

    # the send counts
    counts = comm.allgather(newlength)
    counts = numpy.array(counts, order='C')
    
    # the send offsets
    offsets = numpy.zeros_like(counts, order='C')
    offsets[1:] = counts.cumsum()[:-1]
    
    # do the scatter
    comm.Barrier()
    comm.Scatterv([data, (counts, offsets), dt], [recvbuffer, dt])
    dt.Free()
    return recvbuffer

class EmptyRankType(object):
    def __repr__(self):
        return "EmptyRank"
EmptyRank = EmptyRankType()

class LinearTopology(object):
    """ Helper object for the topology of a distributed array 
    """ 
    def __init__(self, local, comm):
        self.local = local
        self.comm = comm

    def heads(self):
        """
        The first items on each rank. 
        
        Returns
        -------
        heads : list
            a list of first items, EmptyRank is used for empty ranks
        """

        head = EmptyRank
        if len(self.local) > 0:
            head = self.local[0]

        return self.comm.allgather(head)

    def tails(self):
        """
        The last items on each rank. 
        
        Returns
        -------
        tails: list
            a list of last items, EmptyRank is used for empty ranks
        """
        tail = EmptyRank
        if len(self.local) > 0:
            tail = self.local[-1]

        return self.comm.allgather(tail)

    def prev(self):
        """
        The item before the local data.

        This method fetches the last item before the local data.
        If the rank before is empty, the rank before is used. 

        If no item is before this rank, EmptyRank is returned

        Returns
        -------
        prev : scalar
            Item before local data, or EmptyRank if all ranks before this rank is empty.

        """

        tails = [EmptyRank]
        oldtail = EmptyRank
        for tail in self.tails():
            if tail is EmptyRank:
                tails.append(oldtail)
            else:
                tails.append(tail)
                oldtail = tail
        prev = tails[self.comm.rank]
        return prev

    def next(self):
        """
        The item after the local data.

        This method the first item after the local data. 
        If the rank after current rank is empty, 
        item after that rank is used. 

        If no item is after local data, EmptyRank is returned.

        Returns
        -------
        next : scalar
            Item after local data, or EmptyRank if all ranks after this rank is empty.

        """
        heads = []
        oldhead = EmptyRank
        for head in self.heads():
            if head is EmptyRank:
                heads.append(oldhead)
            else:
                heads.append(head)
                oldhead = head
        heads.append(EmptyRank)

        next = heads[self.comm.rank + 1]
        return next
    

class DistributedArray(object):
    """
    Distributed Array Object

    A distributed array is striped along ranks

    Attributes
    ----------
    comm : :py:class:`mpi4py.MPI.Comm`
        the communicator

    local : array_like
        the local data

    """
    def __init__(self, local, comm=MPI.COMM_WORLD):
        self.local = local
        self.comm = comm
        self.topology = LinearTopology(local, comm)

    def sort(self, orderby=None):
        """
        Sort array globally by key orderby.

        Due to a limitation of mpsort, self[orderby] must be u8.

        """
        mpsort.sort(self.local, orderby)

    def __getitem__(self, key):
        return DistributedArray(self.local[key], self.comm)

    def unique_labels(self):
        """
        Assign unique labels to sorted local. 

        .. warning ::

            local data must be sorted, and of simple type. (numpy.unique)

        Returns
        -------
        label   :  :py:class:`DistributedArray`
            the new labels, starting from 0

        """
        prev, next = self.topology.prev(), self.topology.next()
         
        junk, label = numpy.unique(self.local, return_inverse=True)
        if len(self.local) == 0:
            Nunique = 0
        else:
            # watch out: this is to make sure after shifting first 
            # labels on the next rank is the same as my last label
            # when there is a spill-over.
            if next == self.local[-1]:
                Nunique = len(junk) - 1
            else:
                Nunique = len(junk)

        label += sum(self.comm.allgather(Nunique)[:self.comm.rank])
        return DistributedArray(label, self.comm)

    def bincount(self, local=False):
        """
        Assign count numbers from sorted local data.

        .. warning ::

            local data must be sorted, and of integer type. (numpy.bincount)

        Parameters
        ----------
        local : boolean
            if local is True, only count the local array.

        Returns
        -------
        N :  :py:class:`DistributedArray`
            distributed counts array. If items of the same value spans other
            chunks of array, they are added to N as well.

        Examples
        --------
        if the local array is [ (0, 0), (0, 1)], 
        Then the counts array is [ (3, ), (3, 1)]
        """
        prev = self.topology.prev()
        if prev is not EmptyRank:
            offset = prev
            if len(self.local) > 0:
                if prev != self.local[0]:
                    offset = self.local[0]
        else:
            offset = 0

        N = numpy.bincount(self.local - offset)

        if local:
            return N

        heads = self.topology.heads()
        tails = self.topology.tails()

        distN = DistributedArray(N, self.comm)
        headsN, tailsN = distN.topology.heads(), distN.topology.tails()

        if len(N) > 0:
            for i in reversed(range(self.comm.rank)):
                if tails[i] == self.local[0]:
                    N[0] += tailsN[i]
            for i in range(self.comm.rank + 1, self.comm.size):
                if heads[i] == self.local[-1]:
                    N[-1] += headsN[i]

        return DistributedArray(N, self.comm)

def test():
    comm = MPI.COMM_WORLD
    local = numpy.empty((comm.rank), 
            dtype=[('key', 'u8'), ('value', 'u8'), ('rank', 'i8')])
    d = DistributedArray(local)
    local['key'] = numpy.arange(len(local))
    local['value'] = d.comm.rank * 10 + local['key']
    local['rank'] = d.comm.rank

    print(d.topology.heads())

    a = d.comm.allgather(d.local['key'])
    if d.comm.rank == 0:
        print('old', a)

    d.sort('key')
    a = d.comm.allgather(d.local['key'])
    if d.comm.rank == 0:
        print('new', a)

    u = d['key'].unique_labels()
    a = d.comm.allgather(u.local)
    if d.comm.rank == 0:
        print('unique', a)

    N = u.bincount()
    a = d.comm.allgather(N.local)
    if d.comm.rank == 0:
        print('count', a)

    N = u.bincount(local=True)
    a = d.comm.allgather(N)
    if d.comm.rank == 0:
        print('count local', a)

    d['key'].local[:] = u.local
    d.sort('value')

    a = d.comm.allgather(d.local['value'])
    if d.comm.rank == 0:
        print('back', a)

if __name__ == '__main__': 
    test()



from nbodykit.extensionpoints import DataSource, DataStream
import numpy
import logging

logger = logging.getLogger('ShiftedObserved')
         
class ShiftedObserverDataSource(DataSource):
    """
    Class to shift the observer using periodic box data
    and impose redshift-space distortions along the the
    resulting observer's line-of-sight
    """
    plugin_name = "ShiftedObserver"
    
    def __init__(self, datasource, translate, rsd=False):        
        
        # make it an array
        self.translate = numpy.array(self.translate)
    
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "establish an explicit observer (outside the box) for a periodic box"
        
        s.add_argument("datasource", type=DataSource.from_config,
            help="the data to translate in order to impose an explicit observer line-of-sight")
        s.add_argument("translate", type=float, nargs=3,
            help="translate the input data by this vector")
        s.add_argument("rsd", type=bool, 
            help="if `True`, impose redshift distortions along the observer's line-of-sight")
    
    def open(self, defaults={}):
        """
        Overload the `open` function to define the `DataStream` on 
        the `datasource` attribute and to use the decorated `read` function
        """
        stream = DataStream(self.datasource, defaults=defaults)
        stream.read = self.transform(stream.read).__get__(stream)
        return stream
    
    def transform(self, read_func):
        """
        Decorator to transform data output of the `DataStream`
        
        Parameters
        ----------
        read_func : callable
            the original `DataStream.read` function
        """
        def read_wrapper(stream, columns, full=False):
            """
            Read from `datasource`, and return the shifted
            data, optionally with RSD implemented from the shifted location
            """
            # length of the original data
            ncols = len(columns)
    
            # request velocity if we are doing RSD
            if self.rsd and 'Velocity' not in columns:
                columns.append('Velocity')
    
            # read position, redshift, and weights from the stream
            for data in read_func(columns, full=full):
                
                # translate the cartesian coordinates
                if 'Position' in columns:
                    i = columns.index('Position')
                    data[i] = data[i] + self.translate
        
                # add in RSD, along observer LOS
                if self.rsd and 'Position' in columns:
                    i = columns.index('Position')
                    j = columns.index('Velocity')
                    pos = data[i]; vel = data[j]
        
                    # the peculiar velocity
                    rad = numpy.linalg.norm(pos, axis=-1)
                    vpec = numpy.einsum("ij,ij->i", pos, vel) / rad
        
                    # shift by the peculiar velocity along LOS
                    # assuming vpec is normalized appropriately
                    line_of_sight = pos / rad[:,None]
                    pos +=  vpec[:,None] * line_of_sight
                
                yield data[:ncols] # remove velocity if not asked for
        return read_wrapper
        
        
        




from nbodykit.extensionpoints import DataSource
import logging

logger = logging.getLogger('Grid')

class GridDataSource(DataSource):
    """
    Class to read field gridded data from a binary file
    
    Notes
    -----
    * Reading is designed to be done by `GridPainter`, which
      reads gridded quantity straight into the `ParticleMesh`
    """
    plugin_name = "Grid"
    
    def __init__(self, path, BoxSize, dtype='f4'):
        pass
    
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "read gridded field data from a binary file"
        
        s.add_argument("path", type=str, help="the file path to load the data from")
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="the size of the isotropic box, or the sizes of the 3 box dimensions")
        s.add_argument("dtype", type=str, help="data type of binary file to read")




from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest
from .. import examples_dir

class RunCorrAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'corr')
           

@add_run_fixture(__name__, RunCorrAlgorithm, 'PairCountCorrelation')
class TestCrossCorr(unittest.TestCase):
    param_file  = "test_cross_corr.params"
    output_file = "test_corr_cross.dat"
    datasources = ['tpm_1.0000.bin.00', 'fof_ll0.200_1.0000.hdf5']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'corr')

  
@add_run_fixture(__name__, RunCorrAlgorithm, 'PairCountCorrelation')
class TestMWhiteCorr1D(unittest.TestCase):
    param_file  = "test_mwhite_halo_1d.params"
    output_file = "test_corr_mwhite_halo_1d.dat"
    datasources = ['mwhite_halo.fofp']

    def test_exit_code(self):
        asserts.test_exit_code(self)

    def test_exception(self):
        asserts.test_exception(self)
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'corr')
        
@add_run_fixture(__name__, RunCorrAlgorithm, 'PairCountCorrelation')
class TestMWhiteCorr2D(unittest.TestCase):
    param_file  = "test_mwhite_halo_2d.params"
    output_file = "test_corr_mwhite_halo_2d.dat"
    datasources = ['mwhite_halo.fofp']

    def test_exit_code(self):
        asserts.test_exit_code(self)

    def test_exception(self):
        asserts.test_exception(self)
    
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'corr')
        
@add_run_fixture(__name__, RunCorrAlgorithm, 'PairCountCorrelation')
class TestMWhiteCorrPoles(unittest.TestCase):
    param_file  = "test_mwhite_halo_poles.params"
    output_file = "test_corr_mwhite_halo_poles.dat"
    datasources = ['mwhite_halo.fofp']

    def test_exit_code(self):
        asserts.test_exit_code(self)

    def test_exception(self):
        asserts.test_exception(self)
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'corr')
        
@add_run_fixture(__name__, RunCorrAlgorithm, 'FFTCorrelation')
class TestFFTCorr(unittest.TestCase):
    param_file  = "test_fft_corr.params"
    output_file = "test_corr_fft.dat"
    datasources = ['tpm_1.0000.bin.00']

    def test_exit_code(self):
        asserts.test_exit_code(self)

    def test_exception(self):
        asserts.test_exception(self)
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'corr')


import os
import os.path
import glob
import traceback
from argparse import Action, SUPPRESS
from sys import modules
from types import ModuleType

# import the parent modules
import nbodykit.plugins
import nbodykit.plugins.algorithms
import nbodykit.plugins.datasource
import nbodykit.plugins.painter
import nbodykit.plugins.transfer
import nbodykit.plugins.user

def load(filename):
    return load2(filename, qualifiedprefix='nbodykit.plugins.user')

def load_builtins():
    path = os.path.abspath(os.path.dirname(__file__))
    return load2(os.path.join(path, 'plugins'), qualifiedprefix='nbodykit')

def load2(filename, qualifiedprefix='nbodykit.plugins.user'):
    """ load a plugin from filename.

        Parameters
        ----------
        filename : string
            path to the .py file
        qualifiedprefix: string
            a prefix to build a qualified name in sys.modules. This
            is used to load the built-in plugins in nbodykit.plugins

        Returns
        -------
        namespace : dict
            modified global namespace of the plugin script.
    """
    if os.path.isdir(filename):
        root = filename.rstrip('/')
        qualname = '.'.join([qualifiedprefix, os.path.basename(root)])
        if qualname not in modules:
            __import__(qualname)
        module = modules[qualname]
        for filename in sorted(os.listdir(root)):
            fullfilename = os.path.join(root, filename)
            basename = os.path.splitext(os.path.basename(filename))[0]
            add = False
            if os.path.isdir(fullfilename):
                add = True
            if filename.endswith('.py'):
                add = True
            if basename.startswith('__'):
                continue

            if add:
                module.__dict__[basename] = load2(fullfilename, qualifiedprefix=qualname)
        return module
    basename, ext = os.path.splitext(os.path.basename(filename))
    qualname = '.'.join([qualifiedprefix, basename])
    module = ModuleType(qualname)

    with open(filename, 'r') as f:
        code = compile(f.read(), filename, 'exec')
        exec(code, module.__dict__)

    modules[qualname] = module
    return module

def ListPluginsAction(extensionpoint):
    class ListPluginsAction(Action):
        def __init__(self,
                     option_strings,
                     dest=SUPPRESS,
                     default=SUPPRESS,
                     help=None, 
                     nargs=None,
                     metavar=None):
            Action.__init__(self, 
                option_strings=option_strings,
                dest=dest,
                default=default,
                nargs=nargs,
                help=help,
                metavar=metavar)
        
        def __call__(self, parser, namespace, values, option_string=None):
            parser.exit(0, extensionpoint.format_help(*values))
            
    return ListPluginsAction



from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest
from .. import examples_dir

class RunTraceHaloAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'trace-halo')
           

@add_run_fixture(__name__, RunTraceHaloAlgorithm, 'TraceHalo')
class TestTPMSnapshot(unittest.TestCase):
    param_file  = "test_tpmsnapshot.params"
    output_file = "fof_ll0.200_1.0000_ichalo.hdf5"
    datasources = ['tpm_1.0000.bin.00', 'fof_ll0.200_1.0000.labels']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)   
    
    def test_result(self):
        asserts.test_hdf_result(self, "TracedFOFGroups")




#! /usr/bin/env python

import argparse
import logging
import os, sys
from mpi4py import MPI

from nbodykit.extensionpoints import Algorithm, algorithms
from nbodykit.extensionpoints import DataSource, Transfer, Painter
from nbodykit.pluginmanager import ListPluginsAction, load

# configure the logging
rank = MPI.COMM_WORLD.rank
name = MPI.Get_processor_name()

def setup_logging(log_level):
    """
    Set the basic configuration of all loggers
    """
    logging.basicConfig(level=log_level,
                        format='rank %d on %s: '%(rank,name) + \
                                '%(asctime)s %(name)-15s %(levelname)-8s %(message)s',
                        datefmt='%m-%d %H:%M')

class HelpAction(argparse.Action):
    """
    Help action that replicates the behavior of subcommands, 
    i.e., `python prog.py subcommand -h` prints the
    help for the subcommand
    """
    def __init__(self,
                 option_strings,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help=None):
        super(HelpAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        name = getattr(namespace, 'algorithm_name')
        if rank == 0:
            if name is not None:
                print(Algorithm.format_help(name))
            else:
                parser.print_help()
        parser.exit()
        
def main():
    
    valid_algorithms = list(vars(algorithms).keys())

    # initialize the main parser
    desc = "Invoke an `nbodykit` algorithm with the given parameters. \n\n"
    desc += "MPI usage: mpirun -n [n] python nbkit.py ... \n\n"
    desc += "Because MPI standard requires the python interpreter in mpirun commandline. \n"
    kwargs = {}
    kwargs['description'] = desc
    kwargs['fromfile_prefix_chars'] = '@'
    kwargs['formatter_class'] = argparse.ArgumentDefaultsHelpFormatter
    parser = argparse.ArgumentParser("nbkit.py", add_help=False, **kwargs)

    # add the arguments to the parser
    parser.add_argument('algorithm_name', choices=valid_algorithms)
    parser.add_argument('config', type=argparse.FileType(mode='r'), nargs='?', default=None,
                        help='the name of the file to read parameters from using YAML syntax; '
                             'if not provided, stdin is read from')
    parser.add_argument('-h', '--help', action=HelpAction, help='help for a specific algorithm')
    parser.add_argument("-X", type=load, action="append", help="add a directory or file to load plugins from")
    parser.add_argument('-v', '--verbose', action="store_const", dest="log_level", 
                        const=logging.DEBUG, default=logging.INFO, 
                        help="run in 'verbose' mode, with increased logging output")
    
    # help arguments
    parser.add_argument('--list-datasources', nargs='*', action=ListPluginsAction(DataSource), 
        metavar='DataSource', help='list DataSource options')
    parser.add_argument('--list-algorithms',  nargs='*', action=ListPluginsAction(Algorithm), 
        metavar='Algorithm', help='list Algorithm options')
    parser.add_argument('--list-painters',  nargs='*', action=ListPluginsAction(Painter), 
        metavar='Painter', help='list Painter options')
    parser.add_argument('--list-transfers',  nargs='*', action=ListPluginsAction(Transfer), 
        metavar='Transfer', help='list Transfer options')

    # configure printing
    parser.usage = parser.format_usage()[6:-1] + " ... \n"
    if MPI.COMM_WORLD.rank != 0:
        parser._print_message = lambda x, file=None: None

    # parse the command-line
    ns, args = parser.parse_known_args()
    alg_name = ns.algorithm_name
    
    # setup logging
    setup_logging(ns.log_level)
        
    # configuration file passed via -c
    if ns.config is not None:
        # ns.config is a file object
        stream = ns.config.read()
    else:
        if MPI.COMM_WORLD.rank == 0:
            stream = sys.stdin.read()
        else:
            stream = None
        # respect the root rank stdin only;
        # on some systems, the stdin is only redirected to the root rank.
        stream = MPI.COMM_WORLD.bcast(stream)
    
    
    # expand environment variables in the input stream
    stream = os.path.expandvars(stream)
    params, extra = Algorithm.parse_known_yaml(alg_name, stream)
        
    # output is required
    if not hasattr(extra, 'output'):
        raise ValueError("parameter `output` is required in configuration; set to `None` for stdout")
    extra = vars(extra)
    output = extra.pop('output')
    
    # print warning if extra parameters ignored
    if MPI.COMM_WORLD.rank == 0 and len(extra):
        ignored = "[ %s ]" % ", ".join(["'%s'" %k for k in extra.keys()])
        logging.warning("the following keywords to `nbkit.py` have been ignored: %s" %ignored)
            
    # initialize the algorithm and run
    alg_class = getattr(algorithms, alg_name)
    alg = alg_class(**vars(params))

    # run and save
    result = alg.run()
    alg.save(output, result) 
       
if __name__ == '__main__':
    main()



import numpy
import logging
import os
from contextlib import contextmanager
from scipy.interpolate import InterpolatedUnivariateSpline as spline

from nbodykit.extensionpoints import DataSource, Painter, algorithms
from nbodykit.distributedarray import GatherArray

logger = logging.getLogger('FKPCatalog')

class FKPCatalog(object):
    """
    A `DataSource` representing a catalog of tracer objects, 
    designed to be used in analysis similar to that first outlined 
    by Feldman, Kaiser, and Peacock (FKP) 1994 (astro-ph/9304022)
    
    In particular, the `FKPCatalog` uses a catalog of random objects
    to define the mean density of the survey, in addition to the catalog
    of data objects
    
    
    Attributes
    ----------
    data: DataSource
        a `DataSource` that returns the position, weight, etc of the 
        true tracer objects, whose intrinsic clustering is non-zero
    randoms: DataSource
        a `DataSource` that returns the position, weight, etc 
        of a catalog of objects generated randomly to match the
        survey geometry and whose instrinsic clustering is zero
    BoxSize:
        the size of the cartesian box -- the Cartesian coordinates
        of the input objects are computed using the input cosmology,
        and then placed into the box
    mean_coordinate_offset: 
        the average coordinate value in each dimension -- this offset
        is used to return cartesian coordinates translated into the
        domain of [-BoxSize/2, BoxSize/2]
    """
    def __init__(self,  data, 
                        randoms, 
                        BoxSize=None, 
                        BoxPad=0.02, 
                        compute_fkp_weights=False, 
                        P0_fkp=None, 
                        nbar=None, 
                        fsky=None):
        """
        Finalize by performing several steps:
        
            1. if `BoxSize` not provided, infer the value 
               from the Cartesian coordinates of the `data` catalog
            2. compute the mean coordinate offset for each 
               Cartesian dimension -- used to re-center the 
               coordinates to the [-BoxSize/2, BoxSize/2] domain
            3. compute the number density as a function of redshift
               from the `data` and store a spline
        """
        # set the cosmology
        self.cosmo = data.cosmo
        if self.cosmo is None:
            raise ValueError("FKPCatalog requires a cosmology")
        if data.cosmo is not randoms.cosmo:
            raise ValueError("mismatch between cosmology instances of `data` and `randoms` in `FKPCatalog`")
            
        # set the comm
        self.comm = data.comm
        if data.comm is not randoms.comm:
            raise ValueError("mismatch between communicators of `data` and `randoms` in `FKPCatalog`")
        
        # data and randoms datasources
        self.data    = data
        self.randoms = randoms

        # optional configuration
        self.BoxSize             = BoxSize
        self.BoxPad              = BoxPad
        self.compute_fkp_weights = compute_fkp_weights
        self.P0_fkp              = P0_fkp
        self.nbar                = nbar
        self.fsky                = fsky
        
        # default painter
        self.painter = Painter.create('DefaultPainter')

    @property
    def fsky(self):
        """
        The sky area fraction (relative to 4pi str); needed for the volume 
        computation when computing `n(z)`
        """
        try:
            return self._fsky
        except:
            cls = self.__class__.__name__
            raise AttributeError("'%s' object has no attribute 'fsky'" %cls)
            
    @fsky.setter
    def fsky(self, val):
        """
        Set the sky fraction, only if not None
        """
        if val is not None:
            self._fsky = val
        
    @property
    def data(self):
        """
        Update the ``data`` attribute, keeping track of the total number 
        of objects
        """
        try:
            return self._data
        except:
            cls = self.__class__.__name__
            raise AttributeError("'%s' object has no attribute 'data'" %cls)
        
    @data.setter
    def data(self, val):
        """
        Set the data
        """
        if not hasattr(self, '_data'):
            self._data = val
        else:
            # open the new data automatically
            if not self.closed:
                
                # close the old data
                if hasattr(self, 'data_stream'):
                    self.data_stream.close()
                    del self.data_stream
                
                # set and open the new data
                self._data = val
                defaults = {'Redshift':-1., 'Nbar':-1., 'Weight':1.}
                self.data_stream = self.data.open(defaults)
                self.verify_data_size() # verify the size
            
            else:
                self._data = val
    
    def verify_data_size(self):
        """
        Verify the size of the data, setting it if need be
        """            
        # make sure the size is set properly
        try:
            size = self.data.size
        except:
            if hasattr(self, 'data_stream') and not self.data_stream.closed:            
                # compute the total number
                for [Position] in self.data_stream.read(['Position'], full=False):
                    continue
                self.data.size = self.data_stream.nread
                logger.debug("setting `data` size to %d" %self.data.size)
                        
    @property
    def closed(self):
        """
        Return `True` if the catalog has been setup and the
        data and random streams are open
        """
        if not hasattr(self, 'data_stream'):
            return True
        elif self.data_stream.closed:
            return True
            
        if not hasattr(self, 'randoms_stream'):
            return True
        elif self.randoms_stream.closed:
            return True
        
        return False
        
    @property
    def nbar(self):
        """
        A callable function that returns the number density ``n(z)`` 
        as a function of redshift (provided via argument)
        """
        return self._nbar
        
    @nbar.setter
    def nbar(self, val):
        """
        Set the number density as a function of redshift n(z)
        """
        if val is not None:
            if isinstance(val, str) and os.path.exists(val):
                try:
                    d = numpy.loadtxt(val)
                    
                    # columns are z, n(z)
                    if d.shape[1] == 2:
                        self._nbar = spline(d[:,0], d[:,1])
                    # columns are z_min, z_max, z_cen, n(z)
                    elif d.shape[1] == 4:
                        self._nbar = spline(d[:,-2], d[:,-1])
                    self._nbar.need_redshift = True
                except:
                    raise ValueError("cannot initialize `n(z)` spline from file '%s'" %val)
            elif numpy.isscalar(val):
                val = float(val)
                self._nbar = lambda: val # return a constant n(z)
                self._nbar.need_redshift = False
            else:
                raise ValueError("error setting n(z) from input value")
        else:
            self._nbar = None
             
    def _define_box(self, coords_min, coords_max):
        """
        Define the Cartesian box by:
        
            * computing the Cartesian coordinates for all objects
            * setting the `BoxSize` attribute, if not provided
            * computing the coorindate offset needed to center the
              data onto the [-BoxSize/2, BoxSize/2] domain
        """   
        # center the data in the first cartesian quadrant
        delta = abs(coords_max - coords_min)
        self.mean_coordinate_offset = 0.5 * (coords_min + coords_max)
        
        # set the box size automatically
        if self.BoxSize is None:
            delta *= 1.0 + self.BoxPad
            self.BoxSize = numpy.ceil(delta) # round up to nearest integer
        
    def open(self):
        """
        Open the catalog by defining the Cartesian box
        and opening the `data` and `randoms` streams
        """
        # open the streams
        defaults = {'Redshift':-1., 'Nbar':-1., 'Weight':1.}
        self.data_stream = self.data.open(defaults=defaults)
        self.randoms_stream = self.randoms.open(defaults=defaults)
                
        # verify data size
        self.verify_data_size()
    
        # need to compute cartesian min/max
        pos_min = numpy.array([numpy.inf]*3)
        pos_max = numpy.array([-numpy.inf]*3)
        
        # loop over the randoms to make the box
        for [pos] in self.randoms_stream.read(['Position'], full=False):
            if len(pos):
                
                # global min/max of cartesian coordinates
                pos_min = numpy.minimum(pos_min, pos.min(axis=0))
                pos_max = numpy.maximum(pos_max, pos.max(axis=0))

        # set the size, if not set already
        if not hasattr(self.randoms, 'size'):
            self.randoms.size = self.randoms_stream.nread
                
        # gather everything to root
        pos_min   = self.comm.gather(pos_min)
        pos_max   = self.comm.gather(pos_max)
        
        # compute the n(z) of the randoms to get n(z)
        if self.nbar is None:
            fsky = 1. if not hasattr(self, 'fsky') else self.fsky
            nz_computer = algorithms.RedshiftHistogram(self.randoms, fsky=fsky)
            try:
                zbins, z_cen, nz = nz_computer.run()
                self.randoms_nbar = spline(z_cen, nz)
            except DataSource.MissingColumn:
                self.randoms_nbar = None
            except Exception as e:
                logger.warning("exception occurred while computing randoms n(z)")
                self.randoms_nbar = None
        
        # rank 0 setups up the box and computes nbar (if needed)
        if self.comm.rank == 0:
            
            # find the global
            pos_min   = numpy.amin(pos_min, axis=0)
            pos_max   = numpy.amax(pos_max, axis=0)
            
            # setup the box, using randoms to define it
            self._define_box(pos_min, pos_max)
        else:
            self.mean_coordinate_offset = None
            
        # broadcast the results that rank 0 computed
        self.BoxSize                = self.comm.bcast(self.BoxSize)
        self.mean_coordinate_offset = self.comm.bcast(self.mean_coordinate_offset)
        
        if self.comm.rank == 0:
            logger.info("BoxSize = %s" %str(self.BoxSize))
            logger.info("cartesian coordinate range: %s : %s" %(str(pos_min), str(pos_max)))
            logger.info("mean coordinate offset = %s" %str(self.mean_coordinate_offset))
            
    def close(self):
        """
        Close the FKPCatalog by close the `data` and `randoms` streams
        """
        if hasattr(self, 'data_stream'):
            self.data_stream.close()
            del self.data_stream
        if hasattr(self, 'randoms_stream'):
            self.randoms_stream.close()
            del self.randoms_stream
                
    def __enter__ (self):
        if self.closed:
            self.open()
        
    def __exit__ (self, exc_type, exc_value, traceback):
        self.close()
                    
    def read(self, name, columns, full=False):
        """
        Read data from `stream`, which is specified by the `name` argument
        """   
        # check valid columns
        valid = ['Position', 'Weight', 'Nbar']
        if any(col not in valid for col in columns):
            raise DataSource.MissingColumn("valid `columns` to read from FKPCatalog: %s" %str(valid))
                             
        if name == 'data':
            stream = self.data_stream
        elif name == 'randoms':
            stream = self.randoms_stream
        else:
            raise ValueError("stream name for FKPCatalog must be 'data' or 'randoms'")
    
        # read position, redshift, and weights from the stream
        columns0 = ['Position', 'Redshift', 'Weight', 'Nbar']
        for [coords, redshift, weight, nbar] in stream.read(columns0, full=full):
        
            # determine if we have unique redshift, nbar arrays
            default_z = stream.isdefault('Redshift', redshift)
            default_nbar = stream.isdefault('Nbar', nbar)
        
            # recentered cartesian coordinates (between -BoxSize/2 and BoxSize/2)
            pos = coords - self.mean_coordinate_offset
            
            # enforce that position is between (-L/2, L/2)
            lim = (pos < -0.5*self.BoxSize)|(pos > 0.5*self.BoxSize)
            if lim.any():
                out_of_bounds = lim.any(axis=1).sum()
                args = (out_of_bounds, name, self.BoxSize)
                errmsg = ("%d '%s' particles have positions outside of the box when using a BoxSize of %s" %args)
                logger.warning(errmsg)
                logger.warning(("the positions of out-of-bounds particles are periodically wrapped into "
                                 "the box domain -- the resulting behavior is undefined"))

            # number density from redshift
            if self.nbar is not None:
                if self.nbar.need_redshift:
                    if default_z:
                        raise ValueError("`n(z)` calculation requires redshift "
                                         "but '%s' DataSource does not support `Redshift` column" %name)
                    nbar = self.nbar(redshift)
                else:
                    # nbar is a single value
                    nbar = numpy.array(self.nbar())
                    nbar = numpy.lib.stride_tricks.as_strided(nbar, (len(pos), nbar.size), (0, nbar.itemsize))
                    nbar = nbar.squeeze()
                    
            elif default_nbar:
                if default_z:
                    raise ValueError("`n(z)` calculation requires redshift "
                                     "but '%s' DataSource does not support `Redshift` column" %name)
            
                if not hasattr(self, 'fsky'):
                    raise ValueError("computing `n(z)` from 'randoms' DataSource, but no `fsky` "
                                     "provided for volume calculation")
                
                if self.randoms_nbar is None:
                    raise ValueError("`n(z)` calculation requires redshift "
                                     "but 'randoms' DataSource does not support `Redshift` column")
                    
                alpha = 1.*self.data.size/self.randoms.size
                nbar = self.randoms_nbar(redshift) * alpha
        
            elif name == 'randoms':
                alpha = 1.*self.data.size/self.randoms.size
                nbar = nbar * alpha
                
            # update the weights with new FKP
            if self.compute_fkp_weights:
                if self.P0_fkp is None:
                    raise ValueError("if 'compute_fkp_weights' is set, please specify a value for 'P0_fkp'")
                weight = 1. / (1. + nbar*self.P0_fkp)
            
            P = {}
            P['Position'] = pos
            P['Weight']   = weight
            P['Nbar']     = nbar
        
            yield [P[key] for key in columns]
            
    def paint(self, pm):
        """
        Paint the FKP weighted density field: ``data - alpha*randoms`` using
        the input `ParticleMesh`
        
        Parameters
        ----------
        pm : ParticleMesh
            the particle mesh instance to paint the density field to
        
        Returns
        -------
        stats : dict
            a dictionary of FKP statistics, including total number, normalization,
            and shot noise parameters (see equations 13-15 of Beutler et al. 2013)
        """  
        if self.closed:
            raise ValueError("'paint' operation on a closed FKPCatalog")
            
        # setup
        columns = ['Position', 'Weight', 'Nbar']
        stats = {}
        A_ran = A_data = 0.
        S_ran = S_data = 0.
        N_ran = N_data = 0
        
        # clear the density mesh
        pm.clear()
        
        # alpha determined from size of data sources
        alpha = 1.*self.data.size/self.randoms.size
        
        # paint -1.0*alpha*N_randoms
        for [position, weight, nbar] in self.read('randoms', columns):
            Nlocal = self.painter.basepaint(pm, position, -alpha*weight)
            A_ran += (nbar*weight**2).sum()
            N_ran += Nlocal
            S_ran += (weight**2).sum()

        A_ran = self.comm.allreduce(A_ran)
        N_ran = self.comm.allreduce(N_ran)
        S_ran = self.comm.allreduce(S_ran)
        
        if N_ran != self.randoms.size:
            args = (N_ran, self.randoms.size)
            raise ValueError("`size` mismatch when painting: `N_ran` = %d, `randoms.size` = %d" %args)

        # paint the data
        for [position, weight, nbar] in self.read('data', columns):
            Nlocal = self.painter.basepaint(pm, position, weight)
            A_data += (nbar*weight**2).sum()
            N_data += Nlocal 
            S_data += (weight**2).sum()
            
        A_data = self.comm.allreduce(A_data)
        N_data = self.comm.allreduce(N_data)
        S_data = self.comm.allreduce(S_data)
        
        if N_data != self.data.size:
            args = (N_data, self.data.size)
            raise ValueError("`size` mismatch when painting: `N_data` = %d, `data.size` = %d" %args)

        # store the stats (see equations 13-15 of Beutler et al 2013)
        # see equations 13-15 of Beutler et al 2013
        stats['N_data'] = N_data; stats['N_ran'] = N_ran
        stats['A_data'] = A_data; stats['A_ran'] = A_ran
        stats['S_data'] = S_data; stats['S_ran'] = S_ran
        stats['alpha'] = alpha
        
        stats['A_ran'] *= alpha
        stats['S_ran'] *= alpha**2
        stats['shot_noise'] = (S_ran + S_data)/A_ran # the final shot noise estimate for monopole
        
        return stats
    
        



import numpy
import logging
from .stripedfile import StripeFile, DataStorage

__all__ = ['TPMSnapshotFile', 'SnapshotFile']

class TPMSnapshotFile(StripeFile):
    def __init__(self, basename, fid, args={}):
        self.filename = basename + ".%02d" % fid

        with open(self.filename, 'rb') as ff:
            header = numpy.fromfile(ff, dtype='i4', count=7)
        self.header = header
        self.npart = int(header[2])
        self.offset_table = {
            'Position' : (('f4', 3), 7 * 4), 
            'Velocity' : (('f4', 3), 7 * 4 + 12 * self.npart),
            'ID' : (('u8'), 7 * 4 + 12 * self.npart * 2),
        }

    @classmethod
    def create(kls, basename, fid, npart, meta={}):
        filename = basename + ".%02d" % fid
        header = numpy.zeros(7, dtype='i4')
        header[2] = npart
        header[0] = 1
        
        with open(filename, 'wb') as ff:
            header.tofile(ff)
        self = kls(basename, fid)
        return self
 
    def read(self, column, mystart=0, myend=-1):
        
        # return zero-length array, if we don't support column
        if column not in self.offset_table:
            return numpy.empty(0)
            
        dtype, offset = self.offset_table[column]
        dtype = numpy.dtype(dtype)
        if myend == -1:
            myend = self.npart

        with open(self.filename, 'rb') as ff:
            # skip header
            ff.seek(offset, 0)
            # jump to mystart of positions
            ff.seek(mystart * dtype.itemsize, 1)
            return numpy.fromfile(ff, count=myend - mystart, dtype=dtype)

    def write(self, column, mystart, data):
        
        # crash if we don't support column
        if column not in self.offset_table:
            raise ValueError("available columns to write: %s" %str(self.offset_table))
        dtype, offset = self.offset_table[column]
        dtype = numpy.dtype(dtype)
        with open(self.filename, 'rb+') as ff:
            # skip header
            ff.seek(offset, 0)
            # jump to mystart of positions
            ff.seek(mystart * dtype.itemsize, 1)
            return data.astype(dtype.base).tofile(ff)

class GadgetSnapshotFile(StripeFile):
    def __init__(self, basename, fid, args=dict(ptype=1,
        posdtype='f8', veldtype='f4', massdtype='f4', iddtype='u8')):

        self.ptype = args['ptype']
        self.filename = basename + ".%d" % fid
        posdtype = numpy.dtype(args['posdtype'])
        veldtype = numpy.dtype(args['veldtype'])
        massdtype = numpy.dtype(args['massdtype'])
        iddtype = numpy.dtype(args['iddtype'])
        
        with open(self.filename, 'rb') as ff:
            header = numpy.fromfile(ff, dtype='u1', count=256+8)

        headerdtype = [
          ('N', ('u4', 6)), ('mass', ('f8', 6)),
          ('time', 'f8'), ('redshift', 'f8'),
          ('flag_sfr', 'i4'), ('flag_feedback', 'i4'),
          ('Ntot_low', ('u4', 6)), ('flag_cool', 'i4'),
          ('Nfiles', 'i4'), ('boxsize', 'f8'),
          ('OmegaM', 'f8'), ('OmegaL', 'f8'),
          ('h', 'f8'), ('flag_sft', 'i4'),
          ('flag_met', 'i4'), ('Ntot_high', ('u4', 6)),
          ('flag_entropy', 'i4'), ('flag_double', 'i4'),
          ('flag_ic_info', 'i4'), ('flag_lpt_scalingfactor', 'i4'),
          ('flag_pressure_entropy', 'i1'), ('Ndims', 'i1'),
          ('densitykerneltype', 'i1'), ('unused', ('u1', 45)),
        ]
        header = header[4:-4].view(dtype=headerdtype)[0]
        self.header = header
        self.npart = int(header['N'][self.ptype])

        nall = int(header['N'].sum())
        nmass = int((header['N'] * (header['mass'] == 0)).sum())
        nallc = int(header['N'][:self.ptype].sum())
        nmassc = int((header['N'][:self.ptype] * 
                        (header['mass'][:self.ptype] == 0)).sum())
        o = {}
        offset = 256 + 8
        o['Position'] = ((posdtype, 3), offset + 4 + nallc *
                posdtype.itemsize * 3) 
        offset += posdtype.itemsize * 3 * nall + 8
        o['Velocity'] = ((veldtype, 3), offset + 4 + nallc *
                veldtype.itemsize * 3) 
        offset += veldtype.itemsize * 3* nall + 8
        o['ID'] = (iddtype, offset + 4 + nallc * 8)
        offset += iddtype.itemsize * 3 * nall
        o['Mass'] = (massdtype, offset + 4 + nmassc * massdtype.itemsize),
        self.offset_table = o

    def read(self, column, mystart=0, myend=-1):
        # return zero-length array, if we don't support column
        if column not in self.offset_table:
            return numpy.empty(0)
            
        dtype, offset = self.offset_table[column]
        dtype = numpy.dtype(dtype)
        if myend == -1:
            myend = self.npart

        if column == 'Mass' and self.header['mass'][self.ptype] != 0:
            return numpy.ones(myend - mystart, dtype=dtype) * self.header['mass'][self.ptype]

        with open(self.filename, 'rb') as ff:
            # skip header
            ff.seek(offset, 0)
            # jump to mystart of positions
            ff.seek(mystart * dtype.itemsize, 1)
            return numpy.fromfile(ff, count=myend - mystart, dtype=dtype)

class GadgetGroupTabFile(StripeFile):
    def __init__(self, basename, fid, args=dict(
        posdtype='f8', veldtype='f4', massdtype='f4')):

        self.filename = basename + ".%d" % fid
        posdtype = numpy.dtype(args['posdtype'])
        veldtype = numpy.dtype(args['veldtype'])
        massdtype = numpy.dtype(args['massdtype'])

        headerdtype = [('N', ('i4', (1,))),
                ('Ntot', ('i4', (1,))),
                ('Nids',  'i4'),
                ('TotNids', 'u8'),
                ('Nfiles', 'i4')
        ]
        with open(self.filename, 'rb') as ff:
            header = numpy.fromfile(ff, dtype=headerdtype, count=1)[0]

        self.header = header
        self.npart = int(header['N'])

        o = {}
        offset = header.itemsize
        o['Length'] = ('i4', offset)
        offset += 4 * self.npart
        o['Offset'] = ('i4', offset)
        offset += 4 * self.npart
        o['Mass'] =  (massdtype, offset)
        offset += massdtype.itemsize * self.npart
        o['Position'] = ((posdtype, 3), offset)
        offset += posdtype.itemsize * 3 * self.npart
        o['Velocity'] = ((veldtype, 3), offset)
        offset += veldtype.itemsize * 3 * self.npart
        self.offset_table = o

    def read(self, column, mystart=0, myend=-1):
        # return zero-length array, if we don't support column
        if column not in self.offset_table:
            return numpy.empty(0)
            
        dtype, offset = self.offset_table[column]
        dtype = numpy.dtype(dtype)
        if myend == -1:
            myend = self.npart

        with open(self.filename, 'rb') as ff:
            # skip header
            ff.seek(offset, 0)
            # jump to mystart of positions
            ff.seek(mystart * dtype.itemsize, 1)
            return numpy.fromfile(ff, count=myend - mystart, dtype=dtype)

class Snapshot(object):
    def __init__(self, filename, filetype):
        self.npart = numpy.array(
            [ff.npart for ff in filetype.enum(filename)],
            dtype='i8')

        self.filename = filename
        self.filetype = filetype
        if len(self.npart) == 0:
            raise IOError("No files were found under `%s`" % filename)

    @classmethod
    def create(kls, filename, filetype, npart):
        """ create a striped snapshot. 
            npart is a list of npart for each file 
        """
        for fid, npart1 in enumerate(npart):
            filetype.create(filename, fid, npart1)

        self = kls(filename, filetype)
        return self

    def get_file(self, i):
        return self.filetype(self.filename, i)

    def read(self, column, start, end):
        """this function provides a continuous view of multiple files"""

        NpartPerFile = self.npart
        NpartCumFile = numpy.concatenate([[0], numpy.cumsum(self.npart)])
        data = []
        # add an empty item
        ff = self.filetype(self.filename, 0)
        data.append(ff.read(column, 0, 0))

        for i in range(len(NpartPerFile)):
            if end <= NpartCumFile[i]: continue
            if start >= NpartCumFile[i+1]: continue
            # find the intersection in this file
            mystart = max(start - NpartCumFile[i], 0)
            myend = min(end - NpartCumFile[i], NpartPerFile[i])

            ff = self.filetype(self.filename, i)
            data.append(ff.read(column, mystart, myend))
    
        return numpy.concatenate(data, axis=0)

    def write(self, column, start, data):
        """this function provides a continuous view of multiple files"""

        NpartPerFile = self.npart
        NpartCumFile = numpy.concatenate([[0], numpy.cumsum(self.npart)])
        ff = self.filetype(self.filename, 0)
        end = start + len(data)
        offset = 0
        for i in range(len(NpartPerFile)):
            if end <= NpartCumFile[i]: continue
            if start >= NpartCumFile[i+1]: continue
            # find the intersection in this file
            mystart = max(start - NpartCumFile[i], 0)
            myend = min(end - NpartCumFile[i], NpartPerFile[i])

            ff = self.filetype(self.filename, i)
            ff.write(column, mystart, data[offset:offset + myend - mystart])
            offset += myend - mystart
    
        return 

class HaloLabelFile(StripeFile):
    """
    nbodykit halo label file

    Attributes
    ----------
    npart : int
        Number of particles in this file
    linking_length : float
        linking length. For example, 0.2 or 0.168
        
    """
    def __init__(self, filename, fid, args):
        self.filename = filename + ".%02d" % fid
        with open(self.filename, 'rb') as ff:
            self.npart = numpy.fromfile(ff, 'i4', 1)
            self.linking_length = numpy.fromfile(ff, 'f4', 1)
        self.offset_table = {
            'Label': ('i4', 8),
        }
        self.read = TPMSnapshotFile.read.__get__(self)
        self.write = TPMSnapshotFile.write.__get__(self)


def Read2DPlainText(filename):
    """
    Reads the plain text storage of a 2D measurement,
    as output by the `nbodykit.plugins.Measurement2DStorage` plugin
    
    Returns
    -------
    data : dict
        dictionary holding the `edges` data, as well as the
        data columns for the P(k,mu) measurement
    metadata : dict
        any additional metadata to store as part of the 
        P(k,mu) measurement
    """
    toret = {}
    with open(filename, 'r') as ff:
        
        # read number of k and mu bins are first line
        Nk, Nmu = [int(l) for l in ff.readline().split()]
        N = Nk*Nmu
        
        # names of data columns on second line
        columns = ff.readline().split()
        
        lines = ff.readlines()
        data = numpy.array([float(l) for line in lines[:N] for l in line.split()])
        data = data.reshape((Nk, Nmu, -1)) #reshape properly to (Nk, Nmu)
                        
        # store as return dict, making complex arrays from real/imag parts
        i = 0
        while i < len(columns):
            name = columns[i]
            nextname = columns[i+1] if i < len(columns)-1 else ''
            if name.endswith('.real') and nextname.endswith('.imag'):
                name = name.split('.real')[0]
                toret[name] = data[...,i] + 1j*data[...,i+1]
                i += 2
            else:
                toret[name] = data[...,i]
                i += 1
        
        # read the edges for k and mu bins
        edges = []
        l1 = int(lines[N].split()[-1]); N = N+1
        edges.append(numpy.array([float(line) for line in lines[N:N+l1]]))
        l2 = int(lines[N+l1].split()[-1]); N = N+l1+1
        edges.append(numpy.array([float(line) for line in lines[N:N+l2]]))
        toret['edges'] = edges
        
        # read any metadata
        metadata = {}
        if len(lines) > N+l2:
            N_meta = int(lines[N+l2].split()[-1])
            N = N + l2 + 1
            meta = lines[N:N+N_meta]
            for line in meta:
                fields = line.split()
                cast = fields[-1]
                if cast in __builtins__:
                    metadata[fields[0]] = __builtins__[cast](fields[1])
                elif hasattr(numpy, cast):
                     metadata[fields[0]] = getattr(numpy, cast)(fields[1])
                else:
                    raise TypeError("Metadata must have builtin or numpy type")

    return toret, metadata
    
def Read1DPlainText(filename):
    """
    Reads the plain text storage of a 1D measurement,
    as output by the `nbodykit.plugins.Measurement1DStorage` plugin.
    
    Notes
    -----
    *   If `edges` is present in the file, they will be returned
        as part of the metadata, with the key `edges`
    *   If the first line of the file specifies column names, 
        they will be returned as part of the metadata with the
        `columns` key
    
    Returns
    -------
    data : array_like
        the 1D data stacked vertically, such that each columns
        represents a separate data variable
    metadata : dict
        any additional metadata to store as part of the 
        P(k) measurement 
    """
    # data list
    data = []
    
    # extract the metadata
    metadata = {}
    make_float = lambda x: float(x[1:])
    with open(filename, 'r') as ff:
        
        currline = 0
        lines = ff.readlines()
        
        metadata['columns'] = None
        if lines[0][0] == '#':
            try:
                metadata['columns'] = lines[0][1:].split()
            except:
                pass
        
        while True:
            
            # break if we are at the EOF
            if currline == len(lines): break
            line = lines[currline]
            
            if not line: 
                currline += 1
                continue
                
            if line[0] != '#':
                data.append([float(l) for l in line.split()])
            else:
                line = line[1:]
                
                # read edges
                if 'edges' in line:
                    fields = line.split()
                    N = int(fields[-1]) # number of edges
                    metadata['edges'] = numpy.array([make_float(l) for l in lines[currline+1:currline+1+N]])
                    currline += 1+N
                    continue
        
                # read metadata
                if 'metadata' in line:                
                    # read and cast the metadata properly
                    fields = line.split()
                    N = int(fields[-1]) # number of individual metadata lines
                    for i in range(N):
                        fields = lines[currline+1+i][1:].split()
                        cast = fields[-1]
                        if cast in __builtins__:
                            metadata[fields[0]] = __builtins__[cast](fields[1])
                        elif hasattr(numpy, cast):
                             metadata[fields[0]] = getattr(numpy, cast)(fields[1])
                        else:
                            raise TypeError("metadata must have builtin or numpy type")
                    currline += 1+N
                    continue
                
            # add to the data
            currline += 1
            
    return numpy.asarray(data), metadata
                
            
                
        



from nbodykit.extensionpoints import Algorithm
from nbodykit.extensionpoints import DataSource, Transfer, Painter

import os
import numpy
import logging

def FieldType(ns):
    """
    Construct and return a `Field`:
    a tuple of (`DataSource`, `Painter`, `Transfer`)
    
    Notes
    -----
    *   the default Painter is set to `DefaultPainter`
    *   the default Transfer chain is set to 
        [`NormalizeDC`, `RemoveDC`, `AnisotropicCIC`]
    
    Parameters
    ----------
    fields_dict : OrderedDict
        an ordered dictionary where the keys are Plugin names
        and the values are instantiated Plugins
        
    Returns
    -------
    field : list
        list of (DataSource, Painter, Transfer)
    """         
    # define the default Painter and Transfer
    default_painter = Painter.create("DefaultPainter")
    default_transfer = [Transfer.create(x) for x in ['NormalizeDC', 'RemoveDC', 'AnisotropicCIC']]
    
    # start with a default option for (DataSource, Painter, Transfer)
    field = [None, default_painter, default_transfer]
    
    # set the DataSource
    if 'DataSource' not in ns:
        raise ValueError("exactly one `DataSource` per field must be specified")    
    field[0] = getattr(ns, 'DataSource')
    
    # set the Painter
    if 'Painter' in ns:
        field[1] = getattr(ns, 'Painter')
    
    # set the Transfer
    if 'Transfer' in ns:
        field[2] = getattr(ns, 'Transfer')
        
    return field
    
class FFTPowerAlgorithm(Algorithm):
    """
    Algorithm to compute the 1d or 2d power spectrum and multipoles
    in a periodic box, using an FFT
    """
    plugin_name = "FFTPower"
    logger = logging.getLogger(plugin_name)
    
    def __init__(self, mode, Nmesh, field, other=None, los='z', Nmu=5, 
                    dk=None, kmin=0., quiet=False, poles=[], paintbrush='cic'):
                      
        # combine the two fields
        self.fields = [self.field]
        if self.other is not None:
            self.fields.append(self.other)
        
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "periodic power spectrum calculator via FFT"

        # required
        s.add_argument("mode", type=str, choices=['1d', '2d'], 
            help='compute the power as a function of `k` or `k` and `mu`') 
        s.add_argument("Nmesh", type=int, 
            help='the number of cells in the gridded mesh')
        
        # the first field
        s.add_argument('field', type=FieldType,
            help="first data field; a tuple of (DataSource, Painter, Transfer)")
        s.add_argument('field.DataSource', type=DataSource.from_config, required=True,
            help='the 1st DataSource; run `nbkit.py --list-datasources` for all options')
        s.add_argument('field.Painter', type=Painter.from_config, required=False, 
            help='the 1st Painter; run `nbkit.py --list-painters` for all options')
        s.add_argument('field.Transfer', nargs='*', type=Transfer.from_config, required=False,
            help='the 1st Transfer chain; run `nbkit.py --list-transfers` for all options')
        
        # the other field
        s.add_argument('other', type=FieldType, required=False,
            help="the other data field; a tuple of (DataSource, Painter, Transfer)")
        s.add_argument('other.DataSource', type=DataSource.from_config, required=False,
            help='the 2nd DataSource; run `nbkit.py --list-datasources` for all options')
        s.add_argument('other.Painter',  type=Painter.from_config, required=False, 
            help='the 2nd Painter; run `nbkit.py --list-painters` for all options')
        s.add_argument('other.Transfer', nargs='*', type=Transfer.from_config, required=False,
            help='the 2nd Transfer chain; run `nbkit.py --list-transfers` for all options')
            
        # optional
        s.add_argument("los", type=str, choices="xyz",
            help="the line-of-sight direction -- the angle `mu` is defined with respect to")
        s.add_argument("Nmu", type=int,
            help='the number of mu bins to use from mu=[0,1]; if `mode = 1d`, then `Nmu` is set to 1' )
        s.add_argument("dk", type=float,
            help='the spacing of k bins to use; if not provided, the fundamental mode of the box is used')
        s.add_argument("kmin", type=float,
            help='the edge of the first `k` bin to use; default is 0')
        s.add_argument('quiet', type=bool,
            help="silence the logging output")
        s.add_argument('poles', type=int, nargs='*',
            help='if specified, also compute these multipoles from P(k,mu)')
        s.add_argument('paintbrush', type=lambda x: x.lower(), choices=['cic', 'tsc'],
            help='the density assignment kernel to use when painting; '
                 'CIC (2nd order) or TSC (3rd order)')
            
    def run(self):
        """
        Run the algorithm, which computes and returns the power spectrum
        """
        from nbodykit import measurestats
        from pmesh.particlemesh import ParticleMesh
        
        if self.comm.rank == 0: self.logger.info('importing done')
        
        # setup the particle mesh object, taking BoxSize from the painters
        pm = ParticleMesh(self.fields[0][0].BoxSize, self.Nmesh, 
                            paintbrush=self.paintbrush, dtype='f4', comm=self.comm)

        # only need one mu bin if 1d case is requested
        if self.mode == "1d": self.Nmu = 1
    
        # measure
        y3d, stats1, stats2 = measurestats.compute_3d_power(self.fields, pm, comm=self.comm)
        x3d = pm.k
    
        # get the number of objects (in a safe manner)
        N1 = stats1.get('Ntot', -1)
        N2 = stats2.get('Ntot', -1)
        
        # binning in k out to the minimum nyquist frequency 
        # (accounting for possibly anisotropic box)
        dk = 2*numpy.pi/pm.BoxSize.min() if self.dk is None else self.dk
        kedges = numpy.arange(self.kmin, numpy.pi*pm.Nmesh/pm.BoxSize.max() + dk/2, dk)
    
        # project on to the desired basis
        muedges = numpy.linspace(0, 1, self.Nmu+1, endpoint=True)
        edges = [kedges, muedges]
        result, pole_result = measurestats.project_to_basis(pm.comm, x3d, y3d, edges, 
                                                            poles=self.poles, 
                                                            los=self.los, 
                                                            symmetric=True)
                                                            
        # compute the metadata to return
        Lx, Ly, Lz = pm.BoxSize
        meta = {'Lx':Lx, 'Ly':Ly, 'Lz':Lz, 'volume':Lx*Ly*Lz, 'N1':N1, 'N2':N2}                                                    
        
        # return all the necessary results
        return edges, result, pole_result, meta

    def save(self, output, result):
        """
        Save the power spectrum results to the specified output file
        
        Parameters
        ----------
        output : str
            the string specifying the file to save
        result : tuple
            the tuple returned by `run()` -- first argument specifies the bin
            edges and the second is a dictionary holding the data results
        """
        from nbodykit.storage import MeasurementStorage
        
        # only the master rank writes        
        if self.comm.rank == 0:
            
            edges, result, pole_result, meta = result
            if self.mode == "1d":
                cols = ['k', 'power', 'modes']
                result = [numpy.squeeze(result[i]) for i in [0, 2, 3]]
                edges_ = edges[0]
            else:
                edges_ = edges
                cols = ['k', 'mu', 'power', 'modes']
                
            # write binned statistic
            self.logger.info('measurement done; saving result to %s' %output)
            storage = MeasurementStorage.create(self.mode, output)
            storage.write(edges_, cols, result, **meta)
        
            # write multipoles
            if len(self.poles):
                filename, ext = os.path.splitext(output)
                pole_output = filename + '_poles' + ext
            
                # format is k pole_0, pole_1, ...., modes_1d
                self.logger.info('saving ell = %s multipoles to %s' %(",".join(map(str,self.poles)), pole_output))
                storage = MeasurementStorage.create('1d', pole_output)
            
                k, poles, N = pole_result
                cols = ['k'] + ['power_%d' %l for l in self.poles] + ['modes']
                pole_result = [k] + [pole for pole in poles] + [N]
                storage.write(edges[0], cols, pole_result, **meta)
                

class FFTCorrelationAlgorithm(Algorithm):
    """
    Algorithm to compute the 1d or 2d correlation function and multipoles
    in a periodic box. This FFTs the measured power spectrum to compute
    the correlation function
    """
    plugin_name = "FFTCorrelation"
    logger = logging.getLogger(plugin_name)
    
    def __init__(self, mode, Nmesh, field, other=None, los='z', Nmu=5, 
                    dk=None, kmin=0., quiet=False, poles=[], paintbrush='cic'):
            
        # combine the two fields
        self.fields = [self.field]
        if self.other is not None:
            self.fields.append(self.other)

    @classmethod
    def register(cls):
        
        cls.schema.description = "correlation spectrum calculator via FFT in a periodic box"
        for name in FFTPowerAlgorithm.schema:
            cls.schema[name] = FFTPowerAlgorithm.schema[name]
            
    def run(self):
        """
        Run the algorithm, which computes and returns the correlation function
        """
        from nbodykit import measurestats
        from pmesh.particlemesh import ParticleMesh

        if self.comm.rank == 0: self.logger.info('importing done')

        # setup the particle mesh object, taking BoxSize from the painters
        pm = ParticleMesh(self.fields[0][0].BoxSize, self.Nmesh, 
                            paintbrush=self.paintbrush, dtype='f4', comm=self.comm)

        # only need one mu bin if 1d case is requested
        if self.mode == "1d": self.Nmu = 1

        # measure
        y3d, stats1, stats2 = measurestats.compute_3d_corr(self.fields, pm, comm=self.comm)
        x3d = pm.x
        
        # get the number of objects (in a safe manner)
        N1 = stats1.get('Ntot', -1)
        N2 = stats2.get('Ntot', -1)

        # make the bin edges
        dr = pm.BoxSize[0] / pm.Nmesh
        redges = numpy.arange(0, pm.BoxSize[0] + dr * 0.5, dr)

        # project on to the desired basis
        muedges = numpy.linspace(0, 1, self.Nmu+1, endpoint=True)
        edges = [redges, muedges]
        result, pole_result = measurestats.project_to_basis(pm.comm, x3d, y3d, edges,
                                                            poles=self.poles,
                                                            los=self.los,
                                                            symmetric=False)

        # compute the metadata to return
        Lx, Ly, Lz = pm.BoxSize
        meta = {'Lx':Lx, 'Ly':Ly, 'Lz':Lz, 'volume':Lx*Ly*Lz, 'N1':N1, 'N2':N2}

        # return all the necessary results
        return edges, result, pole_result, meta

    def save(self, output, result):
        """
        Save the correlation function results to the specified output file

        Parameters
        ----------
        output : str
            the string specifying the file to save
        result : tuple
            the tuple returned by `run()` -- first argument specifies the bin
            edges and the second is a dictionary holding the data results
        """
        from nbodykit.storage import MeasurementStorage

        # only the master rank writes
        if self.comm.rank == 0:

            edges, result, pole_result, meta = result
            if self.mode == "1d":
                cols = ['r', 'corr', 'modes']
                result = [numpy.squeeze(result[i]) for i in [0, 2, 3]]
                edges_ = edges[0]
            else:
                edges_ = edges
                cols = ['r', 'mu', 'corr', 'modes']

            # write binned statistic
            self.logger.info('measurement done; saving result to %s' %output)
            storage = MeasurementStorage.create(self.mode, output)
            storage.write(edges_, cols, result, **meta)

            # write multipoles
            if len(self.poles):
                filename, ext = os.path.splitext(output)
                pole_output = filename + '_poles' + ext

                # format is k pole_0, pole_1, ...., modes_1d
                self.logger.info('saving ell = %s multipoles to %s' %(",".join(map(str,self.poles)), pole_output))
                storage = MeasurementStorage.create('1d', pole_output)

                k, poles, N = pole_result
                cols = ['k'] + ['power_%d' %l for l in self.poles] + ['modes']
                pole_result = [k] + [pole for pole in poles] + [N]
                storage.write(edges[0], cols, pole_result, **meta)
    





import inspect
import functools
from collections import namedtuple, OrderedDict
import yaml
from argparse import Namespace

class ConfigurationError(Exception):   
    pass

class PluginParsingError(Exception):   
    pass
 
def ordered_load(stream, Loader=yaml.SafeLoader, object_pairs_hook=OrderedDict):
    """
    Load from yaml into OrderedDict to preserve the ordering used 
    by the user
    
    see: http://stackoverflow.com/questions/5121931/
    """
    class OrderedLoader(Loader):
        pass
    def construct_mapping(loader, node):
        loader.flatten_mapping(node)
        return object_pairs_hook(loader.construct_pairs(node))
    OrderedLoader.add_constructor(
        yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
        construct_mapping)
    return yaml.load(stream, OrderedLoader)


def fill_namespace(ns, arg, config, missing):
    """
    Recursively fill a namespace from a configuration file
    such that arguments with subfields get their own namespaces, etc
    """
    name = arg.name.split('.')[-1]
    
    if not len(arg.subfields):  
        if config is not None and name in config:
            value = config.pop(name)
            try:
                setattr(ns, name, ConstructorSchema.cast(arg, value))
            except Exception as e:
                raise ConfigurationError("unable to cast '%s' value: %s" %(arg.name, str(e)))
        else:
            if arg.required:
                missing.append(arg.name)
    else:
        subns = Namespace()
        subconfig = config.pop(name, None)
    
        for k in arg.subfields:
            fill_namespace(subns, arg[k], subconfig, missing)
            
        if len(vars(subns)):
            try:
                setattr(ns, name, ConstructorSchema.cast(arg, subns))
            except Exception as e:
                raise ConfigurationError("unable to cast '%s' value: %s" %(arg.name, str(e)))
            
 
def ReadConfigFile(config_stream, schema):
    """
    Read parameters from a file using YAML syntax
    
    The function uses the specified `schema` to:
        * infer default values
        * check if parameter values are consistent with `choices`
        * infer the `type` of each parameter
        * check if any required parameters are missing
    """
    from nbodykit.cosmology import Cosmology
    from nbodykit.extensionpoints import set_nbkit_cosmo
    from nbodykit.pluginmanager import load

    # make a new namespace
    ns, unknown = Namespace(), Namespace()

    # read the yaml config file
    try:
        config = ordered_load(config_stream)
        if isinstance(config, (str, type(None))):
            raise Exception("no valid keys found")
    except Exception as e:
        raise ConfigurationError("error parsing YAML file: %s" %str(e))
    
    # first load any plugins
    plugins = []
    if 'X' in config:
        plugins = config['X']
        if isinstance(plugins, str):
            plugins = [plugins]
        for plugin in plugins: load(plugin)
        config.pop('X')
    
    # now load cosmology
    cosmo = None
    if 'cosmo' in config:
       cosmo = Cosmology(**config.pop('cosmo'))
       set_nbkit_cosmo(cosmo)
                
    # fill the namespace, recursively 
    missing = []
    extra = config.copy()
    for name in schema:
        fill_namespace(ns, schema[name], extra, missing)
    
    # store any unknown values
    for k in extra:
        setattr(unknown, k, extra[k])
    
    # crash if we don't have all required args
    if len(missing):
        raise ValueError("missing required arguments: %s" %str(missing))
    return ns, unknown

ArgumentBase = namedtuple('Argument', ['name', 'required', 'type', 'default', 'choices', 'nargs', 'help', 'subfields'])
class Argument(ArgumentBase):
    """
    Class to represent an argument in the `ConstructorSchema`
    """
    def __new__(cls, name, required, type=None, default=None, choices=None, nargs=None, help="", subfields=None):
        if subfields is None: subfields = OrderedDict()
        return super(Argument, cls).__new__(cls, name, required, type, default, choices, nargs, help, subfields)

    def __getitem__(self, key):
        if isinstance(key, str):
            return self.subfields[key]
        return ArgumentBase.__getitem__(self, key)

    def _asdict(self):
        # FIXME: the override to getitem seems to be messing up Python 3.
        # is it used at all?
        d = {}
        for f in self._fields:
            d[f] = getattr(self, f)
        return d

class ConstructorSchema(OrderedDict):
    """
    An `OrderedDict` of `Argument` objects, which are `namedtuples`.
    Each `Argument` stores the relevant information of that
    argument, included `type`, `help`, `choices`, etc. 
    
    Each `Argument` also stores a `subfields` attribute, which
    is a new `OrderedDict` of `Arguments`, storing any sub-fields
    
    Notes
    -----
    *   you can test whether a full argument 'name' is in the schema
        with the `contains` function
            >> param_name = 'field.DataSource'
            >> contains = schema.contains(param_name)
            
    *   Arguments that are subfields can be accessed 
        in a sequential dict-like fashion:
            >> subarg = schema['field']['DataSource']
    """  
    Argument = Argument 
              
    def __init__(self, description=""):
        super(ConstructorSchema, self).__init__()
        self.description = description
    
    @staticmethod
    def cast(arg, value):
        """
        Convenience function to cast values based
        on the `type` stored in `schema`
        
        Parameters
        ----------
        arg : Argument
            the `Argument` which gives the relevant metadata
            to properly cast value
        value : 
            the value we are casting, using the `type`
            attribute of `arg`
        """
        if arg.nargs is not None:
            if not isinstance(value, list): value = [value]
        if isinstance(arg.nargs, int) and len(value) != arg.nargs:
            raise ValueError("'%s' requires exactly %d arguments" %(arg.name, arg.nargs))
        if arg.nargs == '+' and len(value) == 0:
            raise ValueError("'%s' requires at least one argument" %arg.name) 
        
        cast = arg.type
        if cast is not None: 
            if arg.nargs is not None:
                value = [cast(v) for v in value]
            else:
                value = cast(value)
        return value
                    
    def contains(self, key):
        """
        Check if the schema contains the full argument name, using
        `.` to represent subfields
        """
        split = key.split('.')
        prefix = split[:-1]; name = split[-1]
        obj = self
        for k in prefix:
            obj = self[k].subfields
        
        return name in obj 
    
    def add_argument(self, name, type=None, default=None, choices=None, nargs=None, help=None, required=False):
        """
        Add an argument to the schema
        """                
        # get the prefix
        split = name.split('.')
        prefix = split[:-1]; suffix = split[-1]
        
        # create default parent Arguments that do not exist
        obj = self
        for i, k in enumerate(prefix):
            if k not in obj:
                obj[k] = Argument('.'.join(prefix[:i+1]), required)
            obj = obj[k].subfields
        
        # add new argument (with empty subfields)
        if not self.contains(name):
            obj[suffix] = Argument(name, required, nargs=nargs, type=type, default=default, 
                                    choices=choices, help=help)
                                    
        # overwrite existing object (copying the subfields)
        else:
            obj[suffix] = obj[suffix]._replace(type=type, default=default, choices=choices, 
                                                help=help, required=required, nargs=nargs)
     
    def _arg_info(self, name, arg, indent=0):
        """
        Internal helper function that returns the info string 
        for one argument, indenting to match a specific level
        
        Format: name: description (default=`default`)
        """
        space = "  " if indent == 0 else "   "*(1+indent)
        info = "%s%-20s %s" %(space, name, arg.help)
        if arg.default is not None:
            info += " (default: %s)" %arg.default
        return info
                     
    def _parse_info(self, name, arg, level):
        """
        Internal function to recursively parse a argument and any
        subfields, returning the full into string
        """
        info = self._arg_info(name, arg, indent=level)
        if not len(arg.subfields):
            return info
            
        for k in arg.subfields:
            v = arg.subfields[k]
            info += '\n'+self._parse_info(k, v, level+1)
             
        return info+'\n'
            
    def format_help(self):
        """
        Return a string giving the help 
        """
        toret = ""
        if getattr(self, 'description', ""):
            toret += self.description + '\n\n'
            
        optional = []; required = []
        for k in self:
            arg = self[k]
            info = self._parse_info(k, arg, level=0)
            if arg.required:
                required.append(info)
            else:
                optional.append(info)
            
        toret += "required arguments:\n%s\n" %('-'*18)
        toret += "\n".join(required)
        toret += "\n\noptional arguments:\n%s\n" %('-'*18)
        toret += "\n".join(optional)
        return toret
        
    __str__ = format_help
        
def attribute(name, **kwargs):
    """
    Declare a class attribute, adding it to the schema attached to 
    the function we are decorating
    """
    def _argument(func):
        if not hasattr(func, 'schema'):
            func.schema = ConstructorSchema()
        func.schema.add_argument(name, **kwargs)
        return func
    return _argument
    
def autoassign(init, attach_comm=True, attach_cosmo=False):
    """
    Verify the schema attached to the input `init` function,
    automatically set the input arguments, and then finally
    call `init`
    
    Parameters
    ----------
    init : callable
        the function we are decorating
    allowed : list, optional
        list of names of additional attributes that are allowed to be 
        auto-assigned if passed to the function -- useful for when
        we are automatically setting the cosmology
    attach_comm : bool, optional
        if `True`, set the `comm` attribute to the return value
        of `get_nbkit_comm`; default: True
    """
    # inspect the function
    attrs, varargs, varkw, defaults = inspect.getargspec(init)
    if defaults is None: defaults = []
    
    allowed = []
    if attach_cosmo:
        if 'cosmo' not in init.schema:
            h = 'the `Cosmology` class relevant for the DataSource'
            init.schema.add_argument("cosmo", default=None, help=h)
        allowed.append('cosmo')
    if attach_comm:
        if 'comm' not in init.schema:
            h = 'the global MPI communicator'
            init.schema.add_argument("comm", default=None, help=h)
        allowed.append('comm')
         
    # verify the schema
    update_schema(init, attrs, defaults, allowed=allowed)
         
    @functools.wraps(init)
    def wrapper(self, *args, **kwargs):
        
        # attach the global communicator
        if attach_comm:
            from nbodykit.extensionpoints import get_nbkit_comm
            self.comm = get_nbkit_comm()

        # attach the global cosmology
        if attach_cosmo:
            from nbodykit.extensionpoints import get_nbkit_cosmo
            self.cosmo = get_nbkit_cosmo()
        
        # handle extra allowed keywords (that aren't in signature)
        for k in allowed:
            if k in kwargs:
                setattr(self, k, kwargs.pop(k))
        
        # handle default values
        for attr, val in zip(reversed(attrs), reversed(defaults)):
            setattr(self, attr, val)
        
        # handle positional arguments
        positional_attrs = attrs[1:]
        posargs = {}            
        for attr, val in zip(positional_attrs, args):
            check_choices(init.schema, attr, val)
            posargs[attr] = val
            setattr(self, attr, val)
    
        # handle varargs
        if varargs:
            remaining_args = args[len(positional_attrs):]
            setattr(self, varargs, remaining_args)            
        
        # handle varkw
        if kwargs:
            for attr,val in kwargs.items():
                check_choices(init.schema, attr, val)
                setattr(self, attr, val)
        
        # call the __init__ to confirm proper initialization
        try:
            return init(self, *args, **kwargs)
        except Exception as e:
            
            # get the error message
            errmsg = get_init_errmsg(init.schema, posargs, kwargs)
            
            # format the total message
            args = (self.__class__.__name__,)
            msg = '\n' + '-'*75 + '\n'
            msg += "error initializing __init__ for '%s':\n" %self.__class__.__name__
            msg += "\t%-25s: '%s'\n" %("original error message", str(e))
            if len(errmsg): msg += "%s\n" %errmsg
            msg += '-'*75 + '\n'
            e.args = (msg, )
            raise
            
    return wrapper
    
def get_init_errmsg(schema, posargs, kwargs):
    """
    Return a reasonable error message, accounting for:
    
        * missing arguments
        * extra arguments
        * duplicated positional + keyword arguments
    """
    errmsg = ""
    
    # check duplicated
    duplicated = list(set(posargs.keys()) & set(kwargs.keys()))
    if len(duplicated):
        s = "duplicated arguments"
        errmsg += "\t%-25s: %s\n" %(s, str(duplicated))
        
    # check for missing arguments
    required = [s for s in schema if schema[s].required]
    missing = []
    for r in required:
        if r not in posargs and r not in kwargs:
            missing.append(r)
    if len(missing):
        s = "missing arguments"
        errmsg += "\t%-25s: %s\n" %(s, str(missing))
    
    # check for extra arguments
    keys = list(set(posargs.keys()) | set(kwargs.keys()))
    extra = []
    for k in keys:
        if k not in schema:
            extra.append(k)
    if len(extra):
        s = "extra arguments"
        errmsg += "\t%-25s: %s\n" %(s, str(extra))
    
    return errmsg
    

def check_choices(schema, attr, val):
    """
    Verify that the input values are consistent
    with the `choices`, using the schema
    """
    if attr in schema:
        arg = schema[attr]
        if arg.choices is not None:
            if val not in arg.choices:
                raise ValueError("valid choices for '%s' are: '%s'" %(arg.name, str(arg.choices)))

def update_schema(func, attrs, defaults, allowed=[]):
    """
    Update the schema, which is attached to `func`,
    using information gather from the function's signature, 
    namely `attrs` and `defaults`
    
    This will update the `required` and `default` values
    of the schema, using the signature of `func`
    
    It also verifies certain aspects of the schema, mostly as a
    consistency check on the developer
    
    The `allowed` list provides the names of the parameters
    not in the function signature that are still allowed, because
    they will be set before the function call
    """
    args = attrs[1:] # ignore self

    # get the required names and default names
    required = args; default_names = []
    if defaults is not None and len(defaults):
        required = args[:-len(defaults)]  
        default_names = args[-len(defaults):]
    
    # loop over the schema arguments
    extra = []; missing = default_names + required
    for name in func.schema:
        a = func.schema[name]

        # infer required and defaults and update them
        d = a._asdict()
        d['required'] = a.name in required
        if a.name in default_names:
            d['default'] = defaults[default_names.index(a.name)]

        func.schema[name] = func.schema.Argument(**d)

        # check for extra and missing
        if a.name not in args and a.name not in allowed:
            extra.append(a.name)
        elif a.name in missing:
            missing.remove(a.name)

    # crash if we are missing or got extra (sanity check)
    if len(missing):
        raise ValueError("missing arguments in schema : %s " %str(missing))
    if len(extra):
        raise ValueError("extra arguments in schema : %s" %str(extra))

    # reorder the schema list to match the function signature
    schema_keys = [k for k in func.schema.keys() if k in args]
    if schema_keys != args:
        new_schema = ConstructorSchema(description=func.schema.description)
        for a in args:
            if a in func.schema:
                new_schema[a] = func.schema[a]
        for p in allowed: 
            if p in func.schema:
                new_schema[p] = func.schema[p]
        func.schema = new_schema
        
    # update the doc with the schema documentation
    if func.__doc__:
        func.__doc__ += "\n\n" + func.schema.format_help()
    else:
        func.__doc__ = func.schema.format_help()



from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest, results
from .. import examples_dir

class RunSubsampleAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'subsample')
           

@add_run_fixture(__name__, RunSubsampleAlgorithm, 'Subsample')
class TestTPMSnapshot(unittest.TestCase):
    param_file  = "test_tpmsnapshot.params"
    output_file = "test_subsample_tpm_1.0000.hdf5"
    datasources = ['tpm_1.0000.bin.00']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self) 
        
    def test_result(self):
        asserts.test_hdf_result(self, "Subsample")
                
@add_run_fixture(__name__, RunSubsampleAlgorithm, 'Subsample')
class TestTPMSnapshotMWhite(unittest.TestCase):
    param_file  = "test_tpmsnapshot_mwhite.params"
    output_file = "test_subsample_tpm_1.0000.mwhite"
    datasources = ['tpm_1.0000.bin.00']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)    

    def test_result(self):
        import numpy
        
        # load the binary data
        this, ref = results.get_result_paths(self.output_file)
        this = numpy.fromfile(this)
        ref = numpy.fromfile(ref)
        
        # assert loaded arrays are cose
        numpy.testing.assert_allclose(this, ref, rtol=1e-5, atol=1e-8)
        



"""
Helper routines for ndarray objects

These are really handy things if numpy had them.

"""
import numpy

def equiv_class(labels, values, op, dense_labels=False, identity=None, minlength=None):
    """
    apply operation to equivalent classes by label, on values

    Parameters ----------
    labels  : array_like
        the label of objects, starting from 0.
    values  : array_like
        the values of objects (len(labels), ...)
    op      : :py:class:`numpy.ufunc`
        the operation to apply
    dense_labels: boolean
        If the labels are already dense (from 0 to Nobjects - 1)
        If False, :py:meth:`numpy.unique` is used to convert
        the labels internally

    Returns
    -------
        the value of each equivalent class

    Examples
    --------
    >>> x = numpy.arange(10)
    >>> print equiv_class(x, x, numpy.fmin, dense_labels=True)
    [0 1 2 3 4 5 6 7 8 9]

    >>> x = numpy.arange(10)
    >>> v = numpy.arange(20).reshape(10, 2)
    >>> x[1] = 0
    >>> print equiv_class(x, 1.0 * v, numpy.fmin, dense_labels=True, identity=numpy.inf)
    [[  0.   1.]
     [ inf  inf]
     [  4.   5.]
     [  6.   7.]
     [  8.   9.]
     [ 10.  11.]
     [ 12.  13.]
     [ 14.  15.]
     [ 16.  17.]
     [ 18.  19.]]

    """
    # dense labels
    if not dense_labels:
        junk, labels = numpy.unique(labels, return_inverse=True)
        del junk
    N = numpy.bincount(labels)
    offsets = numpy.concatenate([[0], N.cumsum()], axis=0)[:-1]
    arg = labels.argsort()
    if identity is None: identity = op.identity
    if minlength is None:
        minlength = len(N)

    result = numpy.empty(minlength, dtype=(values.dtype, values.shape[1:]))
    result[:len(N)] = op.reduceat(values[arg], offsets)

    if (N == 0).any():
        result[:len(N)][N == 0] = identity

    if len(N) < minlength:
        result[len(N):] = identity

    return result

def replacesorted(arr, sorted, b, out=None):
    """
    replace a with corresponding b in arr

    Parameters
    ----------
    arr : array_like
        input array
    sorted   : array_like 
        sorted

    b   : array_like

    out : array_like,
        output array
    Result
    ------
    newarr  : array_like
        arr with a replaced by corresponding b

    Examples
    --------
    >>> print replacesorted(numpy.arange(10), numpy.arange(5), numpy.ones(5))
    [1 1 1 1 1 5 6 7 8 9]

    """
    if out is None:
        out = arr.copy()
    if len(sorted) == 0:
        return out
    ind = sorted.searchsorted(arr)
    ind.clip(0, len(sorted) - 1, out=ind)
    arr = numpy.array(arr)
    found = sorted[ind] == arr
    out[found] = b[ind[found]]
    return out




from nbodykit.extensionpoints import DataSource
import numpy
import logging
from nbodykit.utils import selectionlanguage

logger = logging.getLogger('Pandas')
         
class PandasDataSource(DataSource):
    """
    Class to read data from a plaintext file using
    `pandas.read_csv` or a 'pandas-flavored' `HDF5` file
    using `pandas.read_hdf`
    
    Reading method is guessed from the file type, or
    specified via the `ftype` argument. 

    Notes
    -----
    *   `pandas` must be installed to use
    *   when reading plaintext files, the file is assumed to be space-separated
    *   commented lines must begin with `#`, with all other lines
        providing data values to be read
    *   `names` parameter must be equal to the number of data
        columns, otherwise behavior is undefined
    """
    plugin_name = "Pandas"
    
    def __init__(self, path, names, BoxSize, 
                    usecols=None, poscols=['x','y','z'], velcols=None, 
                    rsd=None, posf=1., velf=1., select=None, ftype='auto'):        
        pass
    
    @classmethod
    def register(cls):
        """
        Fill the attribute schema associated with this class
        """
        s = cls.schema
        s.description = "read data from a plaintext or HDF5 file using Pandas"
        
        s.add_argument("path", type=str,
            help="the file path to load the data from")
        s.add_argument("names", type=str, nargs='+', 
            help="names of columns in text file or name of the data group in hdf5 file")
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="the size of the isotropic box, or the sizes of the 3 box dimensions")
        s.add_argument("usecols", type=str, nargs='*', 
            help="only read these columns from file")
        s.add_argument("poscols", type=str, nargs=3,
            help="names of the position columns")
        s.add_argument("velcols", type=str, nargs=3,
            help="names of the velocity columns")
        s.add_argument("rsd", type=str, choices="xyz", 
            help="direction to do redshift distortion")
        s.add_argument("posf", type=float, 
            help="factor to scale the positions")
        s.add_argument("velf", type=float, 
            help="factor to scale the velocities")
        s.add_argument("select", type=selectionlanguage.Query, 
            help='row selection based on conditions specified as string, i.e., "Mass > 1e14"')
        s.add_argument("ftype", choices=['hdf5', 'text', 'auto'], 
            help='format of the Pandas storage container. auto is to guess from the file name.')
    
    def readall(self):
        """
        Read all available data, returning a dictionary
        
        This provides ``Position`` and optionally ``Velocity`` columns
        """
        try:
            import pandas as pd
        except:
            name = self.__class__.__name__
            raise ImportError("pandas must be installed to use %s" %name)
                
        if self.ftype == 'auto':
            if self.path.endswith('.hdf5'):
                self.ftype = 'hdf5'
            else: 
                self.ftype = 'text'
                
        # read in the hdf5 file using pandas
        if self.ftype == 'hdf5':
            data = pd.read_hdf(self.path, self.names[0], columns=self.usecols)
        # read in the plain text file using pandas
        elif self.ftype == 'text':
            kwargs = {}
            kwargs['comment'] = '#'
            kwargs['names'] = self.names
            kwargs['header'] = None
            kwargs['engine'] = 'c'
            kwargs['delim_whitespace'] = True
            kwargs['usecols'] = self.usecols
            data = pd.read_csv(self.path, **kwargs)

        nobj = len(data)
        
        # select based on input conditions
        if self.select is not None:
            mask = self.select.get_mask(data)
            data = data[mask]
        logger.info("total number of objects selected is %d / %d" % (len(data), nobj))

        toret = {}
        
        # get position and velocity, if we have it
        pos = data[self.poscols].values.astype('f4')
        pos *= self.posf
        if self.velcols is not None or self.rsd is not None:
            vel = data[self.velcols].values.astype('f4')
            vel *= self.velf
            toret['Velocity'] = vel

        # shift position by RSD
        if self.rsd is not None:
            dir = "xyz".index(self.rsd)
            pos[:, dir] += vel[:, dir]
            pos[:, dir] %= self.BoxSize[dir]
        toret['Position'] = pos
        
        return toret




from nbodykit.extensionpoints import DataSource
from nbodykit import files 
import numpy

class TPMSnapshotDataSource(DataSource):
    """
    DataSource to read snapshot files from Martin White's TPM simulations
    """
    plugin_name = "TPMSnapshot"
    
    def __init__(self, path, BoxSize, rsd=None, bunchsize=4*1024*1024):
        
        if self.comm.rank == 0:
            datastorage = files.DataStorage(self.path, files.TPMSnapshotFile)
            size = sum(datastorage.npart)
        else:
            size = None
        self.size = self.comm.bcast(size)

    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "read snapshot files from Martin White's TPM"
        s.add_argument("path", type=str, help="the file path to load the data from")
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="the size of the isotropic box, or the sizes of the 3 box dimensions")
        s.add_argument("rsd", choices="xyz", help="direction to do redshift distortion")
        s.add_argument("bunchsize", type=int, help="number of particles to read per rank in a bunch")

    def parallel_read(self, columns, full=False):
        """ 
        read data in parallel. if Full is True, neglect bunchsize.
        
        This supports `Position`, `Velocity` columns
        """
        Ntot = 0
        # avoid reading Velocity if RSD is not requested.
        # this is only needed for large data like a TPMSnapshot
        # for small Pandas reader etc it doesn't take time to
        # read velocity

        if self.rsd is not None:
            newcolumns = set(columns + ['Velocity'])
        else:
            newcolumns = set(columns)

        if 'Mass' in newcolumns:
            newcolumns.remove('Mass')
        if 'Weight' in newcolumns:
            newcolumns.remove('Weight')

        bunchsize = self.bunchsize
        if full: bunchsize = -1

        if self.comm.rank == 0:
            datastorage = files.DataStorage(self.path, files.TPMSnapshotFile)
        else:
            datastorage = None
        datastorage = self.comm.bcast(datastorage)

        for round, P0 in enumerate(
                datastorage.iter(comm=self.comm, 
                    columns=newcolumns, bunchsize=bunchsize)):
            P = dict(zip(newcolumns, P0))
            if 'Position' in P:
                P['Position'] *= self.BoxSize
            if 'Velocity' in P:
                P['Velocity'] *= self.BoxSize

            if self.rsd is not None:
                dir = "xyz".index(self.rsd)
                P['Position'][:, dir] += P['Velocity'][:, dir]
                P['Position'][:, dir] %= self.BoxSize[dir]

            yield [P.get(key, None) for key in columns]




from nbodykit.extensionpoints import DataSource
from nbodykit import files 
import numpy

class TPMLabel(DataSource):
    """
    DataSource to read file of halo labels (halo id per particle), 
    as generated from Martin White's TPM simulations
    """
    plugin_name = "TPMLabel"
    
    def __init__(self, path, bunchsize=4*1024*1024):
        if self.comm.rank == 0:
            datastorage = files.DataStorage(self.path, files.HaloLabelFile)
        else:
            datastorage = None
        datastorage = self.comm.bcast(datastorage)
        self.size = sum(datastorage.npart)
    
    @classmethod
    def register(cls):
        
        s = cls.schema
        s.description = "read file of halo labels as generated from Martin White's TPM"
        s.add_argument("path", type=str, help="the file path to load the data from")
        s.add_argument("bunchsize", type=int, help="number of particle to read in a bunch")

    def parallel_read(self, columns, full=False):
        """ read data in parallel. if Full is True, neglect bunchsize. """
        Ntot = 0
        # avoid reading Velocity if RSD is not requested.
        # this is only needed for large data like a TPMSnapshot
        # for small Pandas reader etc it doesn't take time to
        # read velocity

        bunchsize = self.bunchsize
        if full: bunchsize = -1

        if self.comm.rank == 0:
            datastorage = files.DataStorage(self.path, files.HaloLabelFile)
        else:
            datastorage = None
        datastorage = self.comm.bcast(datastorage)

        for round, P in enumerate(
                datastorage.iter(comm=self.comm, 
                    columns=columns, bunchsize=bunchsize)):
            P = dict(zip(columns, P))

            yield [P.get(key, None) for key in columns]



from nbodykit.extensionpoints import DataSource
import numpy
import logging
from nbodykit.utils import selectionlanguage

logger = logging.getLogger('FOFGroups')

class FOFDataSource(DataSource):
    """
    Class to read field data from a HDF5 FOFGroup data file

    Notes
    -----
    * `h5py` must be installed to use this data source.
    """
    plugin_name = "FOFGroups"
    
    def __init__(self, path, m0, dataset="FOFGroups", BoxSize=None, rsd=None, select=None):
        
        try:
            import h5py
        except:
            name = self.__class__.__name__
            raise ImportError("h5py must be installed to use '%s' reader" %name)

        BoxSize = numpy.empty(3, dtype='f8')
        if self.comm.rank == 0:
            dataset = h5py.File(self.path, mode='r')[self.dataset]
            BoxSize[:] = dataset.attrs['BoxSize']
            logger.info("Boxsize from file is %s" % str(BoxSize))
        BoxSize = self.comm.bcast(BoxSize)

        if self.BoxSize is None:
            self.BoxSize = BoxSize
        else:
            if self.comm.rank == 0:
                logger.info("Overriding BoxSize as %s" % str(self.BoxSize))
        
    @classmethod
    def register(cls):
        s = cls.schema
        s.description = "read data from a HDF5 FOFGroup file"

        s.add_argument("path", type=str, help="the file path to load the data from")
        s.add_argument("m0", type=float, help="the mass unit")
        s.add_argument("dataset", type=str, help="the name of the dataset in HDF5 file")
        s.add_argument("BoxSize", type=cls.BoxSizeParser,
            help="overide the size of the box; can be a scalar or a three-vector")
        s.add_argument("rsd", type=str, choices="xyz", 
            help="direction to do redshift distortion")
        s.add_argument("select", type=selectionlanguage.Query, 
            help='row selection based on conditions specified as string')
    
    def readall(self):
        import h5py

        dataset = h5py.File(self.path, mode='r')[self.dataset]
        data = dataset[...]

        data2 = numpy.empty(len(data),
            dtype=[
                ('Position', ('f4', 3)),
                ('Velocity', ('f4', 3)),
                ('Mass', 'f4'),
                ('Weight', 'f4'),
                ('Length', 'i4'),
                ('Rank', 'i4'),
                ('LogMass', 'f4')])

        data2['Mass'] = data['Length'] * self.m0
        data2['Weight'] = 1.0
        data2['LogMass'] = numpy.log10(data2['Mass'])
        # get position and velocity, if we have it
        data2['Position'] = data['Position'] * self.BoxSize
        data2['Velocity'] = data['Velocity'] * self.BoxSize
        data2['Rank'] = numpy.arange(len(data))
        # select based on input conditions
        if self.select is not None:
            mask = self.select.get_mask(data2)
            data2 = data2[mask]

        nobj = (len(data2), len(data))

        logger.info("total number of objects selected is %d / %d" % nobj)

        if self.rsd is not None:
            dir = "xyz".index(self.rsd)
            data2['Position'][:, dir] += data2['Velocity'][:, dir]
            data2['Position'][:, dir] %= self.BoxSize[dir]

        return {key: data2[key].copy() for key in data2.dtype.names}




from nbodykit.extensionpoints import Painter
import numpy
import logging

logger = logging.getLogger('MomentumPainter')

class MomentumPainter(Painter):
    plugin_name = "MomentumPainter"
    
    def __init__(self, velocity_component, moment=1):
        self._comp_index = "xyz".index(self.velocity_component)

    @classmethod
    def register(cls):
        s = cls.schema
        s.add_argument("velocity_component", type=str, choices='xyz',
            help="which velocity component to grid, either 'x', 'y', 'z'")
        s.add_argument("moment", type=int, 
            help="the moment of the velocity field to paint, i.e., "
                 "`moment=1` paints density*velocity, `moment=2` paints density*velocity^2")

    def paint(self, pm, datasource):
        """
        Paint the ``DataSource`` specified by ``input`` onto the 
        ``ParticleMesh`` specified by ``pm``
    
        Parameters
        ----------
        pm : ``ParticleMesh``
            particle mesh object that does the painting
        datasource : ``DataSource``
            the data source object representing the field to paint onto the mesh
            
        Returns
        -------
        stats : dict
            dictionary of statistics, usually only containing `Ntot`
        """
        pm.clear()
        stats = {}
        Nlocal = 0
        
        # open the datasource stream (with no defaults)
        with datasource.open() as stream:
        
            # just paint density as usual
            if self.moment == 0: 
                for [position] in stream.read(['Position']):
                    Nlocal += self.basepaint(pm, position)
            # paint density-weighted velocity moments
            else:
                for position, velocity in stream.read(['Position', 'Velocity']):
                    Nlocal += self.basepaint(pm, position, velocity[:,self._comp_index]**self.moment)
    
        # total N
        stats['Ntot'] = self.comm.allreduce(Nlocal)
        
        # normalize config-space velocity field by mean number density
        norm = 1.*stats['Ntot']/pm.BoxSize.prod()
        pm.real[:] /= norm

        return stats
            



#! /usr/bin/env python

import logging
import os
import tempfile
from string import Formatter

from mpi4py import MPI
from nbodykit.extensionpoints import Algorithm, algorithms
from nbodykit.utils.taskmanager import TaskManager

# setup the logging
rank = MPI.COMM_WORLD.rank
name = MPI.Get_processor_name()
logging.basicConfig(level=logging.DEBUG,
                    format='rank %d on %s: '%(rank,name) + \
                            '%(asctime)s %(name)-15s %(levelname)-8s %(message)s',
                    datefmt='%m-%d %H:%M')
logger = logging.getLogger('nbkit-batch')

def replacements_from_file(value):
    """
    Provided an existing file name, read the file into 
    a dictionary. The keys are interpreted as the string format name, 
    and the values are a list of values to iterate over for each job
    """
    if not os.path.exists(value):
        raise RuntimeError("for `replacements_from_file`, file `%s` does not exist" %value)
    
    toret = {}
    with open(value) as f:
        code = compile(f.read(), value, 'exec')
        exec(code, globals(), toret)

    return toret
        
def tasks_parser(value):
    """
    Given a string of the format ``key tasks``, split the string and then
    try to parse the ``tasks``, by first trying to evaluate it, and then
    simply splitting it and interpreting the results as the tasks. 
    
    The general use cases are: 
    
    1) "box: range(2)" -> key = `box`, tasks = `[0, 1]`
    2) "box: ['A', 'B' 'C']" -> key = `box`, tasks = `['A', 'B', 'C']`
    """
    import yaml
    
    try:
        fields = yaml.load(value)
        keys = list(fields.keys())
        if len(fields) != 1:
            raise Exception
    except:
        raise ValueError("specify iteration tasks via the format: ``-i key: [task1, task2]``")
    
    key = keys[0]
    if isinstance(fields[key], list):
        parsed = fields[key]
    else:
        # try to eval into a list
        try:
            parsed = eval(fields[key])
            if not isinstance(parsed, list):
                raise ValueError("result of `eval` on iteration string should be list" %(fields[key]))
        except:
            raise ValueError("tried but failed to `eval` iteration string `%s`" %(fields[key]))

    return [key, parsed]

class BatchAlgorithmDriver(object):
    """
    Class to facilitate running algorithms in batch mode
    """
    def __init__(self, comm, 
                       algorithm_name, 
                       config, 
                       cpus_per_worker, 
                       task_dims, 
                       task_values, 
                       log_level=logging.INFO, 
                       extras={},
                       use_all_cpus=False):
        """
        Parameters
        ----------
        comm : MPI communicator
            the global communicator that will be split and divided
            amongs the independent workers
        algorithm_name : str
            the string name of the `Algorithm` we are running
        config : str
            the name of the file holding the template config file, which
            will be updated for each task that is performed
        cpus_per_worker : int
            the desired number of ranks assigned to each independent
            worker, when iterating over the tasks in parallel
        task_dims : list
            a list of strings specifying the names of the task dimensions -- 
            these specify the string formatting key when updating the config
            template file for each task value
        task_value : list
            a list of tuples specifying the task values which will be iterated 
            over -- each tuple should be the length of `task_dims`
        log_level : int, optional
            an integer specifying the logging level to use -- default
            is the `INFO` level
        extras : dict, optional
            a dictionary where the values are lists of string replacements, with
            length equal to the total number of tasks -- if the keys are present
            in the config file, the string formatting will update the config
            file with the `ith` element of the list for the `ith` iteration
        use_all_cpus : bool, optional
            if `True`, then do not force each worker group to have 
            exactly `cpus_per_worker` ranks, instead including the remainder 
            as well; default is `False`
        """
        logger.setLevel(log_level)
        
        self.algorithm_name  = algorithm_name
        self.algorithm_class = getattr(algorithms, algorithm_name) 
        self.template        = os.path.expandvars(open(config, 'r').read())
        self.cpus_per_worker = cpus_per_worker
        self.task_dims       = task_dims
        self.task_values     = task_values
        self.extras          = extras
        
        self.comm      = comm
        self.size      = comm.size
        self.rank      = comm.rank
        
        # initialize the worker pool
        kws = {'comm':self.comm, 'use_all_cpus':use_all_cpus}
        if log_level <= logging.DEBUG: kws['debug'] = True
        self.workers = TaskManager(self.compute_one_task, self.cpus_per_worker, **kws)
                
    @classmethod
    def create(cls, comm=None, desc=None):
        """
        Parse the task manager and return the ``BatchAlgorithmDriver`` instance
        """
        import inspect 
        
        if comm is None: comm = MPI.COMM_WORLD
        args_dict = cls.parse_args(desc)
        args_dict['comm'] = comm
        
        # inspect the __init__ function
        args, varargs, varkw, defaults = inspect.getargspec(cls.__init__)
        
        # determine the required arguments
        args = args[1:] # remove 'self'
        if defaults:
            required = args[:-len(defaults)]
        else:
            required = args
            
        # get the args, kwargs to pass to __init__
        fargs = tuple(args_dict[p] for p in required)
        fkwargs = {}
        if defaults:
            for i, p in enumerate(defaults):
                name = args[-len(defaults)+i]
                fkwargs[name] = args_dict.get(name, defaults[i])
        
        return cls(*fargs, **fkwargs)
        
    @classmethod
    def parse_args(cls, desc=None):
        """
        Parse command-line arguments that are needed to initialize a 
        `BatchAlgorithmDriver` class
        
        Parameters
        ----------
        desc : str, optional
            the description of to use for this parser
        """
        import argparse
        import itertools
        
        # parse
        parser = argparse.ArgumentParser(description=desc) 
        
        # first argument is the algorithm name
        h = 'the name of the `Algorithm` to run in batch mode'
        valid_algorithms = list(vars(algorithms).keys())  
        parser.add_argument(dest='algorithm_name', choices=valid_algorithms, help=h)  
        
        # the number of independent workers
        h = """the desired number of ranks assigned to each independent
                worker, when iterating over the tasks in parallel""" 
        parser.add_argument('cpus_per_worker', type=int, help=h)
    
        # now do the required named arguments
        required_named = parser.add_argument_group('required named arguments')
        
        # specify the tasks along one dimension 
        h =  """given a string of the format ``key: tasks``, split the string and then
                try to parse the ``tasks``, by first trying to evaluate it, and then
                simply splitting it and interpreting the results as the tasks. 
        
                The general use cases are: 
        
                1) "box: range(2)" -> key = `box`, tasks = `[0, 1]`
                2) "box: [A, B, C]" -> key = `box`, tasks = `['A', 'B', 'C']`
                
                If multiple options passed with `-i` flag, then the total tasks to 
                perform will be the product of the tasks lists passed"""
        required_named.add_argument('-i', dest='tasks', action='append', 
                type=tasks_parser, required=True, help=h)
    
        # the template config file
        h = """the name of the template config file (using YAML synatx) that 
                provides the `Algorithm` parameters; the file should use 
                ``string.format`` syntax to indicate which variables will be 
                updated for each task, i.e., an input file could be specified 
                as 'input/DataSource_box{box}.dat', if `box` were one of the task 
                dimensions"""
        required_named.add_argument('-c', '--config', type=str, help=h)
    
        # read any extra string replacements from file
        h = """file providing extra string replaces, with lines of the form 
                 `tag = ['tag1', 'tag2']`; if the keys match keywords in the 
                 template param file, the file with be updated with
                 the `ith` value for the `ith` task"""
        parser.add_argument('--extras', dest='extras', default={}, type=replacements_from_file, help=h)
    
        h = "set the logging output to debug, with lots more info printed"
        parser.add_argument('--debug', help=h, action="store_const", dest="log_level", 
                            const=logging.DEBUG, default=logging.INFO)
                            
        h = "if `True`, include all available cpus in the worker pool"
        parser.add_argument('--use_all_cpus', action='store_true', help=h)
                                
        args = parser.parse_args()
        
        # format the tasks, taking the product of multiple task lists
        keys = []; values = []
        for [key, tasks] in args.tasks:
            keys.append(key)
            values.append(tasks)

        # take the product
        if len(keys) > 1:
            values = list(itertools.product(*values))
        else:
            values = values[0]
            
        # save
        args.task_dims = keys
        args.task_values = values
        
        return vars(args)
            
    def compute(self):
        """
        Compute all of the tasks
        """             
        # compute the work
        results = self.workers.compute(self.task_values)
        return results
            
    def compute_one_task(self, itask, task):
        """
        Run the algorithm once, using the parameters specified by `task`,
        which is the `itask` iteration
    
        Parameters
        ----------
        itask : int
            the integer index of this task
        task : tuple
            a tuple of values representing this task value
        """
        # if you are the pool's root, write out the temporary parameter file
        this_config = None
        if self.workers.subcomm.rank == 0:
            
            # extract the keywords that we need to format from template file
            formatter = Formatter()
            kwargs = [kw for _, kw, _, _ in formatter.parse(self.template) if kw]
                
            # initialize a temporary file
            with tempfile.NamedTemporaryFile(delete=False) as ff:
                
                this_config = ff.name
                logger.debug("creating temporary file: %s" %this_config)
                
                # key/values for this task 
                if len(self.task_dims) == 1:
                    possible_kwargs = {self.task_dims[0] : task}
                else:
                    possible_kwargs = dict(zip(self.task_dims, task))
                    
                # any extra key/value pairs for this tasks
                if self.extras is not None:
                    for k in self.extras:
                        possible_kwargs[k] = self.extras[k][itask]
                        
                # do the string formatting if the key is present in template
                valid = {k:possible_kwargs[k] for k in possible_kwargs if k in kwargs}
                ff.write(self.template.format(**valid).encode())
        
        # bcast the file name to all in the worker pool
        this_config = self.workers.subcomm.bcast(this_config, root=0)

        # configuration file passed via -c
        params, extra = Algorithm.parse_known_yaml(self.algorithm_name, open(this_config, 'r').read())
        
        # output is required
        output = getattr(extra, 'output', None)
        if output is None:
            raise ValueError("argument `output` is required in config file")
            
        # initialize the algorithm and run
        alg = self.algorithm_class(**vars(params))
        result = alg.run()
        alg.save(output, result)

        # remove temporary files
        if self.workers.subcomm.rank == 0:
            if os.path.exists(this_config): 
                logger.debug("removing temporary file: %s" %this_config)
                os.remove(this_config)
                
        return 0

   
if __name__ == '__main__' :
    
    desc = """iterate (possibly in parallel) over a set of configuration parameters, 
              running the specified `Algorithm` for each"""
    batch = BatchAlgorithmDriver.create(MPI.COMM_WORLD, desc=desc)
    batch.compute()
    
    



from .pipeline import RunAlgorithm, add_run_fixture
from . import os, asserts, unittest, pytest
from .. import examples_dir

# import halotools for HOD power tests
try: import halotools; missing_halotools=False
except: missing_halotools = True
    
class RunPowerAlgorithm(RunAlgorithm):
    run_dir = os.path.join(examples_dir, 'power')
           

@add_run_fixture(__name__, RunPowerAlgorithm, 'BianchiFFTPower')
class TestBianchi1(unittest.TestCase):
    param_file  = "test_bianchi_1.params"
    output_file = "test_power_bianchi_1.dat"
    datasources = ['test_bianchi_data.dat', 'test_bianchi_randoms.dat']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)   
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'power') 


@add_run_fixture(__name__, RunPowerAlgorithm, 'BianchiFFTPower')
class TestBianchi2(unittest.TestCase):
    param_file  = "test_bianchi_2.params"
    output_file = "test_power_bianchi_2.dat"
    datasources = ['test_bianchi_data.dat']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)    
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'power')


@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestCrossMomentum(unittest.TestCase):
    param_file  = "test_cross_momentum.params"
    output_file = "test_power_cross_momentum.dat"
    datasources = ['fastpm_1.0000', 'fof_ll0.200_1.0000.hdf5']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)  
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'power')


@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestCrossPower(unittest.TestCase):
    param_file  = "test_cross_power.params"
    output_file = "test_power_cross.dat"
    datasources = ['fastpm_1.0000', 'fof_ll0.200_1.0000.hdf5']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)   
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'power')
         

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestFastPM1D(unittest.TestCase):
    param_file  = "test_fastpm_1d.params"
    output_file = "test_power_fastpm_1d.dat"
    datasources = ['fastpm_1.0000']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)   
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'power')
         

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestFastPM2D(unittest.TestCase):
    param_file  = "test_fastpm_2d.params"
    output_file = "test_power_fastpm_2d.dat"
    datasources = ['fastpm_1.0000']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)   
    
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'power')
         

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestFOFGroups(unittest.TestCase):
    param_file  = "test_fofgroups.params"
    output_file = "test_power_fofgroups.dat"
    datasources = ['fof_ll0.200_1.0000.hdf5']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)    
    
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'power')
         

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestPandasHDF(unittest.TestCase):
    param_file  = "test_pandas_hdf.params"
    output_file = "test_power_pandas_hdf.dat"
    datasources = ['pandas_data.hdf5']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)    
         
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'power')

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestPandasPlaintext(unittest.TestCase):
    param_file  = "test_pandas_plaintext.params"
    output_file = "test_power_pandas_plaintext.dat"
    datasources = ['plaintext_data.txt']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self) 
    
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'power')
         

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestPlaintext(unittest.TestCase):
    param_file  = "test_plaintext.params"
    output_file = "test_power_plaintext.dat"
    datasources = ['plaintext_data.txt']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self) 
    
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'power')

         

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestTPMSnapshot1D(unittest.TestCase):
    param_file  = "test_tpmsnapshot_1d.params"
    output_file = "test_power_tpmsnapshot_1d.dat"
    datasources = ['tpm_1.0000.bin.00']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)   
    
    def test_result(self):
        asserts.test_dataset_result(self, '1d', 'power')

         

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestTPMSnapshot2D(unittest.TestCase):
    param_file  = "test_tpmsnapshot_2d.params"
    output_file = "test_power_tpmsnapshot_2d.dat"
    datasources = ['tpm_1.0000.bin.00']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self) 
    
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'power')
         

@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestMWhiteHalo(unittest.TestCase):
    param_file  = "test_mwhite_halo.params"
    output_file = "test_power_mwhite_halo.dat"
    datasources = ['mwhite_halo.fofp']
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)
    
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'power')

@pytest.mark.skipif(missing_halotools, reason="requires `halotools` package")
@add_run_fixture(__name__, RunPowerAlgorithm, 'FFTPower')
class TestZhengHOD(unittest.TestCase):
    param_file  = "test_zheng_hod.params"
    output_file = "test_power_zheng_hod.dat"
    datasources = []
    
    def test_exit_code(self):
        asserts.test_exit_code(self)
    
    def test_exception(self):
        asserts.test_exception(self)
    
    def test_result(self):
        asserts.test_dataset_result(self, '2d', 'power')